"""Includes ModelSelectionUserInputService class."""
from typing import List

import pandas as pd

from organon.fl.core.helpers import list_helper
from organon.ml.common.enums.target_type import TargetType

from organon.ml.common.helpers.user_input_service_helper import get_enum
from organon.ml.modelling.algorithms.core.abstractions.base_modeller import BaseModeller
from organon.ml.modelling.algorithms.core.enums.modeller import Modeller
from organon.ml.modelling.algorithms.core.enums.modeller_type import ModellerType
from organon.ml.modelling.model_selection.services.user_settings.user_hp_optimization_settings import \
    UserHPOptimizationSettings
from organon.ml.modelling.model_selection.services.user_settings.user_selection_settings import UserSelectionSettings
from organon.ml.modelling.model_selection.settings.enums.search_type import SearchType
from organon.ml.modelling.model_selection.settings.objects.cross_validation_settings import CrossValidationSettings
from organon.ml.modelling.model_selection.settings.objects.hp_optimization_settings import HPOptimizationSettings
from organon.ml.modelling.model_selection.settings.objects.selection_settings import SelectionSettings


class ModelSelectionUserInputService:
    """User input service for model_selection module"""

    @classmethod
    def get_selection_settings(cls, user_settings: UserSelectionSettings) -> SelectionSettings:
        """Generate SelectionSettings from UserSelectionSettings"""
        target_type = get_enum(user_settings.target_type, TargetType)
        return SelectionSettings(user_settings.modellers,
                                 user_settings.train_data, user_settings.target_data,
                                 target_type=target_type, cv_count=user_settings.cv_count,
                                 test_data=user_settings.test_data, test_target_data=user_settings.test_target_data,
                                 add_stacking=user_settings.add_stacking, add_voting=user_settings.add_voting,
                                 num_threads_for_parallel_cv=user_settings.num_threads_for_parallel_cv,
                                 num_threads_for_parallel_fit=user_settings.num_threads_for_parallel_fit)

    @classmethod
    def get_cross_validation_settings(cls, modellers: List[BaseModeller], train_data: pd.DataFrame,
                                      target_data: pd.Series, cv_count: int,
                                      bin_count, return_test_fold) -> CrossValidationSettings:
        """Generate CrossValidationSettings from settings entered by user"""
        if list_helper.is_null_or_empty(modellers):
            raise ValueError("Modellers not given")
        if train_data is None or len(train_data) == 0:
            raise ValueError("train_data cannot be empty")
        if cv_count <= 1:
            raise ValueError("cv_count should be higher than 1")

        first_modeller_type = modellers[0].modeller_type
        if first_modeller_type not in [ModellerType.CLASSIFIER, ModellerType.REGRESSOR]:
            raise ValueError("Modellers should be either classifiers or regressors")

        for modeller in modellers:
            if modeller.modeller_type != first_modeller_type:
                raise ValueError("All modellers should be of same type (classifiers or regressors)")

        if bin_count is not None and first_modeller_type != ModellerType.REGRESSOR:
            raise ValueError("bin_count is used only in regression modelling")

        return CrossValidationSettings(modellers, train_data, target_data, cv_count, bin_count, return_test_fold)

    @classmethod
    def get_hp_optimization_settings(cls, settings: UserHPOptimizationSettings) -> HPOptimizationSettings:
        """Validates settings entered by user and generates settings object for HPOService"""
        if settings.train_data is None or len(settings.train_data) == 0:
            raise ValueError("train_data should not be empty")
        if settings.target_data is None or len(settings.target_data) == 0:
            raise ValueError("target_data should not be empty")

        search_method = get_enum(settings.search_method, SearchType)

        if settings.cv_fold is None and settings.test_data is None:
            raise ValueError("Either cv_fold or test_data should be given")

        if sum(1 for data in [settings.test_data, settings.test_target_data] if data is None) == 1:
            raise ValueError("Both test_data and test_target_data should be given")

        modellers = settings.modellers
        if settings.modellers is not None:
            modellers = [get_enum(modeller, Modeller) for modeller in settings.modellers]
            if [modeller for modeller in modellers if modeller.get_modeller_type() != modellers[0].get_modeller_type()]:
                raise ValueError("Either all modellers should be classifiers or all of them should be regressors")

        if settings.modellers is not None and settings.modeller_params is not None:
            if len(settings.modeller_params) != len(settings.modellers):
                raise ValueError("Please enter parameters for all modellers. "
                                 "Enter an empty dictionary if no parameters will be added for a modeller")

        return HPOptimizationSettings(
            train_data=settings.train_data,
            target_data=settings.target_data,
            modellers=modellers,
            modeller_params=settings.modeller_params,
            search_method=search_method,
            cv_fold=settings.cv_fold,
            scoring_metrics=settings.scoring_metrics,
            test_data=settings.test_data,
            test_target_data=settings.test_target_data,
            search_params=settings.search_params)



"""Includes helper functions for validation"""
from organon.ml.common.enums.target_type import TargetType
from organon.ml.modelling.algorithms.core.enums.modeller_type import ModellerType


def validate_get_target_type(modeller_type: ModellerType, target_type: TargetType):
    """Validates target type for given modeller type"""
    if modeller_type == ModellerType.CLASSIFIER:
        if target_type is None:
            raise ValueError("Target type should be given for classification")
        if target_type not in [TargetType.BINARY, TargetType.MULTICLASS]:
            raise ValueError("Only BINARY and MULTICLASS are accepted as valid target types in classification")
    elif modeller_type == ModellerType.REGRESSOR:
        if target_type is not None and target_type != TargetType.SCALAR:
            raise ValueError("Only SCALAR is accepted as valid target type in regression")
        target_type = TargetType.SCALAR
    return target_type



"""Includes helper methods for transaction source read."""
from typing import List, Dict

import pandas as pd

from organon.afe.core.businessobjects.afe_static_objects import AfeStaticObjects
from organon.afe.dataaccess.services.transaction_file_record_reader import TransactionFileRecordReader
from organon.afe.domain.settings.afe_algorithm_settings import AfeAlgorithmSettings
from organon.afe.domain.settings.afe_data_settings import AfeDataSettings
from organon.afe.domain.settings.feature_generation_settings import FeatureGenerationSettings
from organon.afe.domain.settings.record_source import RecordSource
from organon.fl.core.businessobjects.date_interval import DateInterval


def get_trx_file_record_reader_for_modelling(
        dimension_columns: List[str], quantity_columns: List[str],
        data_source_settings: AfeDataSettings,
        algorithm_settings: AfeAlgorithmSettings,
        interval_per_entity_per_date_col: Dict[str, Dict[str, DateInterval]],
        unique_entities_list: List[str]):
    """Generates TransactionFileRecordReader instance for modelling using given values"""
    distinct_entities_df = pd.DataFrame({AfeStaticObjects.distinct_entities_entity_column_name: unique_entities_list})
    distinct_entities_record_source = RecordSource(source=distinct_entities_df)

    trx_descriptor = data_source_settings.trx_descriptor

    date_columns = __get_date_columns(trx_descriptor.feature_gen_setting)

    min_max_dates_per_date_col = {}
    date_col = trx_descriptor.feature_gen_setting.date_column
    max_obs_date = trx_descriptor.feature_gen_setting.max_observation_date
    if date_col is not None and max_obs_date is not None:
        min_max_dates_per_date_col[date_col.column_name] = (None, max_obs_date)

    trx_reader = TransactionFileRecordReader(
        data_source_settings.trx_descriptor.modelling_raw_input_source,
        date_columns,
        data_source_settings.trx_descriptor.entity_column_name,
        interval_per_entity_per_date_column=interval_per_entity_per_date_col,
        distinct_entity_source=distinct_entities_record_source,
        entity_source_entity_column=AfeStaticObjects.distinct_entities_entity_column_name,
        dimension_columns=dimension_columns, quantity_columns=quantity_columns,
        min_max_dates_per_date_column=min_max_dates_per_date_col,
        max_number_of_transaction_samples=data_source_settings.max_number_of_transaction_samples,
        reading_settings=trx_descriptor.reading_settings,
        dimension_compression_ratio=algorithm_settings.dimension_compression_ratio,
        dimension_max_cardinality=algorithm_settings.dimension_max_cardinality
    )

    return trx_reader


def __get_date_columns(feature_generation_settings: FeatureGenerationSettings):
    date_columns = {}
    if feature_generation_settings.date_column is not None:
        date_columns[feature_generation_settings.date_column.column_name] = \
            feature_generation_settings.date_column
    return date_columns



"""Includes LassoRegressor class"""
from sklearn.linear_model import Lasso

from organon.ml.modelling.algorithms.core.abstractions.base_sklearn_regressor import BaseSklearnRegressor


class LassoRegressor(BaseSklearnRegressor[Lasso]):
    """Wrapper for sklearn lasso modeller"""

    @property
    def regressor_class(self):
        return Lasso



# pylint: skip-file

# cython: cdivision=True
# cython: boundscheck=False
# cython: wraparound=False
# cython: language_level=3
# cython: profile=True

from cython.parallel import prange
import numpy as np
cimport numpy as np
from libc.math cimport isnan, sqrt, NAN
from organon.fl.mathematics.sweep.matrices.i_matrix import IMatrix


ctypedef np.uint8_t uint8


def update_row_k(float[:, :] matrix,
                 float[:] tmp_row,
                 float[:] tmp_col,
                 uint8[:] swept_ever,
                 int k,
                 int n_threads):
    """Updates row for new added feature"""
    cdef int index

    for index in prange(swept_ever.shape[0], nogil=True, num_threads=n_threads):
        if not swept_ever[index]:
            matrix[k, index] = tmp_row[index]
            matrix[index, k] = tmp_col[index]



def sweep_fast_row_c(np.ndarray[float, ndim=2] np_matrix, int length, uint8[:] swept_ever, int row, int n_threads):
    """Does the sweep algorithm along a row"""
    cdef:
        float[:, :] matrix = np_matrix
        float diag_value = matrix[row, row]
        int i, j
        float i_value
    np_matrix[row, :] /= -diag_value

    for i in prange(matrix.shape[0], nogil=True, num_threads=n_threads):
        if i == row:
            continue
        i_value = matrix[i, row]
        # If feature i is swept then we have all the row
        if swept_ever[i]:
            with gil:
                np_matrix[i, :] += i_value * np_matrix[row, :]
            # MklFunctions.daxpy(length, i_value, np.asarray(matrix[row, :]), 1, np.asarray(matrix[i, :]), 1)
        # If not then we have to do operations along the swept indices
        else:
            for j in range(length):
                if swept_ever[j]:
                    matrix[i, j] += i_value * matrix[row, j]
            matrix[i, i] += i_value * matrix[row, i]

        matrix[i, row] = i_value / diag_value
    matrix[row, row] = 1.0 / diag_value

def get_correlation_matrix_c(float[:, :] covariance_matrix, int n_threads) -> np.ndarray:
    """Gets correlation matrix from covariance matrix"""
    cdef:
        float[:, :] matrix = covariance_matrix
        int rows = matrix.shape[0]
        int cols = matrix.shape[1]
        int i, j
        bint nan_i, nan_j, zero_i, zero_j
        float[:, :] sigma = np.empty(shape=(rows, cols), dtype=np.single)

    # This loop checks the covariance matrix. If it can calculate some index of correlation matrix then calculates.
    for i in prange(rows, nogil=True, num_threads=n_threads):
        nan_i = isnan(matrix[i, i])
        zero_i = matrix[i, i] <= 0

        sigma[i, i] = NAN if nan_i else 1.0

        for j in range(i, cols):
            nan_j = isnan(matrix[j, j])
            zero_j = matrix[j, j] <= 0

            if isnan(matrix[i, j]) or nan_i or nan_j:
                sigma[i, j] = NAN

            elif zero_i or zero_j:
                sigma[i, j] = 0.0

            else:
                sigma[i, j] = matrix[i, j] / sqrt(matrix[i, i] * matrix[j, j])

            sigma[j, i] = sigma[i, j]

    return np.asarray(sigma)


def compute_covariances_c(matrix: IMatrix,
                               float[:] means,
                               float[:, :] sigma,
                               float[:, :] corr_matrix,
                               uint8[:] is_computed,
                               int variable,
                               int bin_size,
                               int n_threads):
    """Computes covariances of @variable and other elements
    This method computes covariances JIT. Since we split main
     frame to bins this computes all the bins for a variable.
     As can see from https://en.wikipedia.org/wiki/Covariance
     covariance is normalized dot product of vectors(divided by length)
     Using these covariances we compute the correlations."""

    cdef:
        float[:, :] covariances
        np.ndarray[float, ndim=2] norm_col_variable
        np.ndarray[float, ndim=2] norm_col_i
        long long[:] col_ptr = matrix.col_ptr
        long long[:] row_ptr = matrix.row_ptr
        int i, j, start, end
        bint nan_i, nan_j, zero_i, zero_j
        np.ndarray[float, ndim=2] np_covariances = np.empty(shape=(matrix.shape[1], bin_size), dtype=np.single)

    # First get normalized feature[variable]. Since we have bins get all of them.
    start = variable * bin_size
    end = start + bin_size
    norm_col_variable = get_columns_normalized(matrix.shape[0], start, end, col_ptr, row_ptr, means, n_threads)

    # We need to compute normalization of all features. However normalized feature cannot be sparse anymore.
    # For example: [1, 0, 0, 0] -> [3/4, -1/4, -1/4, -1/4]
    # As you can see there is no zero value in output feature. Therefore we need dense matrix for all features.
    # However we cannot put all matrix into dense matrix due to memory constraints.
    # We split matrix to batches and put these features to dense matrix and compute the dot product of these features
    # via matrix multiplication.
    for i in range(0, matrix.shape[1], bin_size):
        norm_col_i = get_columns_normalized(matrix.shape[0], i, i + bin_size, col_ptr, row_ptr, means, n_threads)
        np_covariances[i:i+bin_size] = np.dot(norm_col_i.T, norm_col_variable)

    # Divide by length
    np_covariances /= (matrix.shape[0] - 1)
    # Get memory view of numpy array to use in nogil section.
    covariances = np_covariances
    # For loop over variables to be computed
    for i in range(start, end):
        # For variable i calculate correlation[i, j] can be done parallel
        for j in prange(is_computed.shape[0], nogil=True, num_threads=n_threads):
            if is_computed[j]:
                continue
            if not isnan(sigma[i, j]):
                continue

            sigma[i, j] = covariances[j, i-start]
            sigma[j, i] = covariances[j, i-start]

            nan_i = isnan(sigma[i, i])
            zero_i = sigma[i, i] <= 0

            if i == j:
                corr_matrix[i, i] = NAN if nan_i else 1.0
            else:
                nan_j = isnan(sigma[j, j])
                zero_j = sigma[j, j] <= 0

                if isnan(sigma[i, j]) or nan_i or nan_j:
                    corr_matrix[i, j] = NAN

                elif zero_i or zero_j:
                    corr_matrix[i, j] = 0.0

                else:
                    corr_matrix[i, j] = sigma[i, j] / sqrt(sigma[i, i] * sigma[j, j])

                corr_matrix[j, i] = corr_matrix[i, j]
        is_computed[i] = True

cdef np.ndarray[float, ndim=2] get_columns_normalized(int row,
                                                      int col_start,
                                                      int col_end,
                                                      long long[:] col_ptr,
                                                      long long[:] row_ptr,
                                                      float[:] means,
                                                      int n_threads):
    """Convert sparse matrix columns to dense matrix"""
    cdef:
        np.ndarray[float, ndim=2] dense_matrix = np.zeros(shape=(row, col_end-col_start), dtype=np.single)
        int start, end, i, j
    # To understand this conversion, see CSC matrix format. Basically slice columns from col_start to col_end
    for j in prange(col_start, col_end, nogil=True, num_threads=n_threads):
        start = col_ptr[j]
        end = col_ptr[j + 1]

        for i in range(start, end):
            # Since we split binary bins. We now that all data is 1.0
            dense_matrix[row_ptr[i], j-col_start] = 1.0

    # Normalize columns.
    dense_matrix -= means.base[col_start:col_end]

    return dense_matrix




"""Includes FormattingHelper class."""
from organon.fl.core.helpers import float_helper


class FormattingHelper:
    """Includes helper methods for value formatting"""

    @staticmethod
    def to_nullable(value: float):
        """Return None if value is an extreme float value, else return value as it is"""
        if float_helper.is_extreme(value):
            return None
        return value



"""
This module includes AfeExecutionHelper class.
"""
from datetime import datetime
from typing import Union, List, Dict, Optional, Tuple, TypeVar, Generic

import pandas as pd

import organon
from organon.afe.core.businessobjects.afe_static_objects import AfeStaticObjects
from organon.afe.domain.common.persist_helper import PersistHelper
from organon.afe.domain.enums.afe_date_column_type import AfeDateColumnType
from organon.afe.domain.enums.afe_learning_type import AfeLearningType
from organon.afe.domain.enums.afe_operator import AfeOperator
from organon.afe.domain.enums.afe_target_column_type import AfeTargetColumnType
from organon.afe.domain.enums.date_resolution import DateResolution
from organon.afe.domain.modelling.supervised.multi_target_modelling_service import MultiTargetModellingService
from organon.afe.domain.modelling.unsupervised.unsupervised_feature_extraction_service import \
    UnsupervisedFeatureExtractionService
from organon.afe.domain.reporting.afe_matplotlib_report_helper import AfeMatplotlibReportHelper
from organon.afe.domain.reporting.afe_plotly_report_helper import AfePlotlyReportHelper
from organon.afe.domain.reporting.base_afe_model_output import BaseAfeModelOutput
from organon.afe.domain.reporting.base_report_helper import BaseReportHelper
from organon.afe.domain.scoring.afe_scoring_service import AfeScoringService
from organon.afe.domain.settings.afe_algorithm_settings import AfeAlgorithmSettings
from organon.afe.domain.settings.afe_modelling_settings import AfeModellingSettings
from organon.afe.domain.settings.afe_output_settings import AfeOutputSettings
from organon.afe.domain.settings.afe_reading_settings import AfeDataReadingSettings
from organon.afe.domain.settings.afe_scoring_settings import AfeScoringSettings
from organon.afe.domain.settings.afe_settings_reader import AfeSettingsReader
from organon.afe.domain.settings.afe_settings_validator import AfeSettingsValidator
from organon.afe.domain.settings.auto_column_decider_settings import AutoColumnDeciderSettings
from organon.afe.domain.settings.feature_generation_settings import FeatureGenerationSettings
from organon.afe.domain.settings.record_source import RecordSource
from organon.afe.domain.settings.target_descriptor import TargetDescriptor
from organon.afe.domain.settings.temporal_grid import TemporalGrid
from organon.afe.domain.settings.trx_descriptor import TrxDescriptor
from organon.afe.services.afe_application_operations import AfeApplicationOperations
from organon.common.helpers import dev_mode_helper
from organon.fl.core.exceptionhandling.known_exception import KnownException
from organon.fl.core.helpers import process_info_helper
from organon.fl.logging.helpers.log_helper import LogHelper
from organon.fl.serialization.serialization_helper import serialize_to_file, deserialize_from_file

AfeModelOutputType = TypeVar("AfeModelOutputType", bound=BaseAfeModelOutput)


class AFE(Generic[AfeModelOutputType]):
    """
    AfeExecutionHelper class
    """

    def __init__(self):
        self._initialize_afe()
        self.model_output: Optional[AfeModelOutputType] = None
        self._settings_reader = self._get_afe_settings_reader()

    @classmethod
    def init_dev_mode(cls, log_to_console: bool = True, log_file: str = "application.log"):
        """Initializes development mode."""
        dev_mode_helper.init_dev_mode(organon.afe.__name__, log_to_console=log_to_console, log_file=log_file)

    @classmethod
    def _initialize_afe(cls):
        AfeApplicationOperations.initialize_app()

    @classmethod
    def _get_afe_settings_reader(cls):
        return AfeSettingsReader()

    @classmethod
    def _get_afe_settings_validator(cls):
        return AfeSettingsValidator()

    def fit(self, settings: AfeModellingSettings) -> AfeModelOutputType:
        """
        Runs Automated Feature Extraction modelling.

        :param settings: AfeModellingSettings instance
        :return: AfeModelOutputType instance
        """
        if not isinstance(settings, AfeModellingSettings):
            raise ValueError("Please give an AfeModellingSettings.")

        modelling_input = settings

        self.model_output = self._run_afe_modelling(modelling_input)

        LogHelper.info("Afe modelling finished successfully.")
        return self.model_output

    def transform(self, trx_source: Union[pd.DataFrame, str],
                  target_source: Union[pd.DataFrame, str] = None,
                  for_date: str or int or datetime = None,
                  num_threads: int = None,
                  trx_reading_settings: Union[AfeDataReadingSettings, Dict] = None,
                  **kwargs):
        """
        Runs Automated Feature Extraction scoring.

        :param trx_source: Transaction record source
        :param target_source: Target records to score
        :param for_date: If given, scoring will be done for only this date
        :param num_threads: Number of threads to use for scoring(cpu_count-1 by default)
        :param trx_reading_settings: Reading settings for trx_source
        :return: Score dataframe
        """
        # pylint: disable=too-many-arguments,unused-argument
        self._check_model_output()
        return self.transform_model(self.model_output, trx_source, target_source=target_source,
                                    for_date=for_date,
                                    num_threads=num_threads,
                                    trx_reading_settings=trx_reading_settings)

    def persist_model(self, location: str):
        """
        Saves model output to given location.

        :param location: file path or db table object.
        """
        self._check_model_output()
        self.persist_model_output(self.model_output, location)

    def load_model(self, location: str, secret_key: str = None, **kwargs):
        """
        Loads model output from given location

        :param location: file path
        :param secret_key: secret key generated while saving model
        """
        # pylint:disable=unused-argument
        self.model_output = self.get_model(location, secret_key=secret_key, **kwargs)

    def show_afe_report(self, use_lib="matplotlib"):
        """
        Shows afe report dashboard.

        :param use_lib: Visualization library to use. Valid values: "matplotlib", "plotly"
        """
        self._check_model_output()
        self.create_afe_report(self.model_output, use_lib=use_lib)

    def show_feature_report(self, use_lib="matplotlib"):
        """
        Shows afe report dashboard.

        :param use_lib: Visualization library to use. Valid values: "matplotlib", "plotly"
        """
        self._check_model_output()
        self.create_feature_report(self.model_output, use_lib=use_lib)

    def get_features_data_frame(self) -> pd.DataFrame:
        """
        Returns afe features as dataframe.

        :return:
        """
        self._check_model_output()
        return self.create_features_data_frame(self.model_output)

    def get_all_features_data_frame(self) -> pd.DataFrame:
        """
        Returns all generated afe features as dataframe.

        :return:
        """
        self._check_model_output()
        return self.create_all_features_data_frame(self.model_output)

    def show_extended_feature_report(self, use_lib="matplotlib"):
        """
        Shows extended feature report.

        :param use_lib: Visualization library to use. Valid values: "matplotlib", "plotly"
        """
        self._check_model_output()
        self.create_extended_feature_report(self.model_output, use_lib=use_lib)

    def show_feature_importance_report(self, target_source_index: int = None,
                                       first_x_features: int = None, use_lib="matplotlib"):
        """
        Shows importance of afe features generated for given target.

        :param first_x_features: Number of columns to show in plot
        :param target_source_index: Index of target(In modelling target_record_source_list)
        :param use_lib: Visualization library to use. Valid values: "matplotlib", "plotly"
        """
        self._check_model_output()
        self.create_feature_importance_report(self.model_output, target_source_index, first_x_features, use_lib=use_lib)

    def get_feature_importance_for_targets(self) -> List[pd.DataFrame]:
        """
        Returns importance of afe features generated for all targets.

        :return: List of pandas dataframes storing feature importance data for targets
        """
        self._check_model_output()
        return self.generate_feature_importance_report_for_targets(self.model_output)

    def get_auc_score_for_targets(self, target_index: int = None) -> List[Tuple[int, float]]:
        """
        Returns auc score for all targets.
        :param target_index: Index of target(In modelling target_record_source_list)
        :return: List of float storing auc score for targets
        """
        self._check_model_output()
        return self.generate_auc_score_for_targets(self.model_output, target_index)

    def show_resource_usage_report(self, use_lib="matplotlib"):
        """
        Shows resource usage dashboard.

        :param use_lib: Visualization library to use. Valid values: "matplotlib", "plotly"
        """
        self._check_model_output()
        self.create_resource_usage_report(self.model_output, use_lib=use_lib)

    def show_roc_curve_report(self, target_source_index: int = None, target_class=None,
                              use_lib="matplotlib"):
        """
        Shows roc_curve for given target.

        :param target_source_index: Index of target(In modelling target_record_source_list)
        :param target_class: target class to report roc_curve for (if multiclass afe modelling)
        :param use_lib: Visualization library to use. Valid values: "matplotlib", "plotly"
        """
        self._check_model_output()
        self.create_roc_curve_report(self.model_output, target_source_index, target_class=target_class, use_lib=use_lib)

    def _check_model_output(self):
        if self.model_output is None:
            raise KnownException("No model has been generated before")

    @classmethod
    def transform_model(cls, model_output: AfeModelOutputType, trx_source: Union[pd.DataFrame, str],
                        target_source: Union[pd.DataFrame, str] = None,
                        for_date: str or int or datetime = None,
                        num_threads: int = None,
                        trx_reading_settings: Union[AfeDataReadingSettings, Dict] = None, **kwargs):
        """
        Runs Automated Feature Extraction scoring for given model output.

        :param model_output: AfeModelOutputType instance to score
        :param trx_source: Transaction record source
        :param target_source: Target records to score
        :param for_date: If given, scoring will be done for only this date
        :param num_threads: Number of threads to use for scoring(cpu_count-1 by default)
        :param trx_reading_settings: Reading settings for trx_source
        :return: Score dataframe
        """
        # pylint: disable=too-many-arguments,unused-argument
        scoring_settings = AfeScoringSettings()

        scoring_settings.model_output = model_output
        if target_source is not None:
            scoring_settings.target_record_source = RecordSource(source=target_source)
        else:
            if for_date is None:
                raise KnownException("Please enter either target_source or for_date parameters to transform.")
        scoring_settings.raw_input_source = RecordSource(source=trx_source)

        if trx_reading_settings is not None:
            scoring_settings.trx_reading_settings = trx_reading_settings \
                if isinstance(trx_reading_settings, AfeDataReadingSettings) \
                else AfeDataReadingSettings(**trx_reading_settings)
        score_df = cls._run_afe_scoring(scoring_settings, num_threads=num_threads, for_date=for_date)
        return score_df

    @classmethod
    def _run_afe_modelling(cls, settings: AfeModellingSettings):
        """Runs afe modelling with given settings"""
        cls._get_afe_settings_validator().validate_modelling_settings(settings)
        cls._get_afe_settings_reader().assign_default_values(settings)

        if settings.afe_learning_type == AfeLearningType.Supervised:
            service = MultiTargetModellingService(settings)
        elif settings.afe_learning_type == AfeLearningType.Unsupervised:
            service = UnsupervisedFeatureExtractionService(settings)
        else:
            raise NotImplementedError

        return service.execute()

    @classmethod
    def _run_afe_scoring(cls, settings: AfeScoringSettings, num_threads: int = None, for_date=None, **kwargs) \
            -> Optional[pd.DataFrame]:
        """Runs afe scoring with given settings"""
        # pylint: disable=unused-argument

        cls._get_afe_settings_validator().validate_scoring_settings(settings)
        cls._get_afe_settings_reader().assign_default_values_for_scoring(settings)

        service = AfeScoringService(settings)

        return service.score(num_threads, for_date=for_date)

    @classmethod
    def get_feature_generation_setting(cls,
                                       temporal_grids: List[Union[TemporalGrid, Dict]] = None,
                                       date_resolution: Union[DateResolution, str] = None,
                                       date_column_name: str = None,
                                       date_column_type: Union[str, AfeDateColumnType] = None,
                                       date_column_format: str = None,
                                       dimension_columns: List[str] = None,
                                       quantity_columns: List[str] = None,
                                       included_operators: List[Union[AfeOperator, str]] = None,
                                       date_offset: int = None,
                                       max_observation_date: datetime = None,
                                       **kwargs
                                       ):
        """
        Generates FeatureGenerationSettings instance.

        :param temporal_grids:
        :param date_resolution:
        :param date_column_name:
        :param date_column_type:
        :param date_column_format:
        :param dimension_columns:
        :param quantity_columns:
        :param included_operators:
        :param date_offset:
        :param max_observation_date:
        :return:
        """
        # pylint: disable=too-many-arguments, unused-argument
        return cls._get_afe_settings_reader().get_feature_generation_setting(
            temporal_grids=temporal_grids,
            date_resolution=date_resolution,
            date_column_name=date_column_name,
            date_column_format=date_column_format,
            date_column_type=date_column_type,
            dimension_columns=dimension_columns,
            quantity_columns=quantity_columns,
            included_operators=included_operators,
            date_offset=date_offset,
            max_observation_date=max_observation_date
        )

    @classmethod
    def get_auto_column_decider_setting(cls, sampling_ratio: float = None,
                                        numeric_to_dimension: int = None,
                                        dimension_distinct_cut_off: float = None,
                                        use_dimension_columns: bool = None,
                                        rejected_dimension_columns: List[str] = None,
                                        use_quantity_columns: bool = None,
                                        rejected_quantity_columns: List[str] = None):
        """
        Generates AutoColumnDeciderSettings instance.

        :param sampling_ratio:
        :param numeric_to_dimension:
        :param dimension_distinct_cut_off:
        :param use_dimension_columns:
        :param rejected_dimension_columns:
        :param use_quantity_columns:
        :param rejected_quantity_columns:
        :return:
        """
        # pylint: disable=too-many-arguments, unused-argument
        return cls._get_afe_settings_reader().get_auto_column_decider_settings(
            sampling_ratio=sampling_ratio,
            numeric_to_dimension=numeric_to_dimension,
            dimension_distinct_cut_off=dimension_distinct_cut_off,
            use_dimension_columns=use_dimension_columns,
            rejected_dimension_columns=rejected_dimension_columns,
            use_quantity_columns=use_quantity_columns,
            rejected_quantity_columns=rejected_quantity_columns
        )

    @classmethod
    def get_trx_descriptor(cls,
                           source: Union[pd.DataFrame, str],
                           entity_column_name: str,
                           feature_generation_settings: Union[FeatureGenerationSettings, Dict],
                           number_of_rows_per_step: int = None) -> TrxDescriptor:
        """
        Generates TrxDescriptor instance for AfeModelling.

        :param source: Transaction record source
        :param entity_column_name: Name of column storing entity ids
        :param feature_generation_settings: Feature generation settings
        :param number_of_rows_per_step: Size of chunks to read in memory while reading target record source
        :return: Transaction record source descriptor
        """
        return cls._get_afe_settings_reader().get_trx_descriptor(
            source=source,
            entity_column_name=entity_column_name,
            feature_gen_setting=feature_generation_settings,
            number_of_rows_per_step=number_of_rows_per_step
        )

    @classmethod
    def get_target_descriptor(cls,
                              entity_column_name: str,
                              date_column_name: str = None,
                              date_column_type: Union[str, AfeDateColumnType] = None,
                              date_column_format: str = None,
                              date_column_db_format: str = None,
                              target_column_name: str = None,
                              target_column_type: Union[str, AfeTargetColumnType] = None,
                              target_positive_category: str = None,
                              target_negative_category: str = None,
                              target_indeterminate_category: str = None,
                              target_exclusion_category: str = None,
                              default_measurement_date: datetime = None,
                              number_of_rows_per_step: int = None) -> TargetDescriptor:
        """
        Generates TargetDescriptor instance for AfeModelling.


        :param target_column_name:
        :param target_column_type:
        :param target_positive_category:
        :param target_negative_category:
        :param target_indeterminate_category:
        :param target_exclusion_category:
        :param entity_column_name: Name of column storing entity ids
        :param date_column_name: Name of column storing event dates
        :param date_column_type:
        :param date_column_format:
        :param date_column_db_format:
        :param default_measurement_date:
        :param number_of_rows_per_step: Size of chunks to read in memory while reading target record source
        :return: Target record source descriptor
        """
        # pylint: disable=too-many-arguments
        return cls._get_afe_settings_reader().get_target_descriptor(
            entity_column_name=entity_column_name,
            date_column_name=date_column_name,
            date_column_type=date_column_type,
            date_column_format=date_column_format,
            date_column_db_format=date_column_db_format,
            target_column_name=target_column_name,
            target_column_type=target_column_type,
            target_positive_category=target_positive_category,
            target_negative_category=target_negative_category,
            target_indeterminate_category=target_indeterminate_category,
            target_exclusion_category=target_exclusion_category,
            default_measurement_date=default_measurement_date,
            number_of_rows_per_step=number_of_rows_per_step)

    @classmethod
    def get_supervised_algorithm_settings(cls,
                                          training_percentage: float = None,
                                          min_data_in_leaf_and_sample_size_control_ratio: float = None,
                                          dimension_compression_ratio: float = None,
                                          dimension_max_cardinality: int = None,
                                          reduction_coverage: float = None,
                                          model_params: Dict = None,
                                          model_fit_params: Dict = None,
                                          final_reduction_coverage: float = None,
                                          final_model_params: Dict = None,
                                          final_model_fit_params: Dict = None,
                                          ):
        """
        Generates AfeSupervisedAlgorithmSettings instance.

        :param training_percentage: percentage of data to be used as training data
        :param min_data_in_leaf_and_sample_size_control_ratio:
        :param dimension_compression_ratio:
        :param dimension_max_cardinality:
        :param reduction_coverage:
        :param model_params:
        :param model_fit_params:
        :param final_reduction_coverage:
        :param final_model_params:
        :param final_model_fit_params:
        :return:
        """
        # pylint: disable=too-many-arguments
        return cls._get_afe_settings_reader().get_supervised_algorithm_settings(
            training_percentage=training_percentage,
            min_data_in_leaf_and_sample_size_control_ratio=min_data_in_leaf_and_sample_size_control_ratio,
            dimension_compression_ratio=dimension_compression_ratio,
            dimension_max_cardinality=dimension_max_cardinality,
            model_params=model_params,
            model_fit_params=model_fit_params,
            reduction_coverage=reduction_coverage,
            final_model_params=final_model_params,
            final_model_fit_params=final_model_fit_params,
            final_reduction_coverage=final_reduction_coverage
        )

    @classmethod
    def get_unsupervised_algorithm_settings(cls,
                                            bin_count: int = None,
                                            r_factor: float = None,
                                            max_column_count: int = None,
                                            is_logging: bool = None,
                                            dimension_compression_ratio: float = None,
                                            dimension_max_cardinality: int = None
                                            ):
        """
        Returns AfeUnsupervisedAlgorithmSettings instance.

        :param bin_count:
        :param r_factor:
        :param max_column_count:
        :param is_logging:
        :param dimension_compression_ratio:
        :param dimension_max_cardinality:
        :return:
        """
        # pylint: disable=too-many-arguments
        return cls._get_afe_settings_reader().get_unsupervised_algorithm_settings(
            bin_count=bin_count,
            r_factor=r_factor,
            max_column_count=max_column_count,
            is_logging=is_logging,
            dimension_compression_ratio=dimension_compression_ratio,
            dimension_max_cardinality=dimension_max_cardinality
        )

    @classmethod
    def get_output_settings(cls,
                            output_folder: str = None,
                            output_prefix: str = None,
                            feature_name_prefix: str = None,
                            enable_feature_lookup_output_to_csv: bool = None,
                            enable_write_output: bool = None,
                            return_all_afe_columns: bool = None,
                            enable_all_feature_lookup_output_to_csv: bool = None,
                            **kwargs):
        """
        Generates AfeOutputSettings instance

        :param output_folder:
        :param output_prefix:
        :param feature_name_prefix:
        :param enable_feature_lookup_output_to_csv:
        :param enable_write_output:
        :return:
        """
        # pylint: disable=too-many-arguments,unused-argument
        return cls._get_afe_settings_reader().get_output_settings(
            output_folder=output_folder, output_prefix=output_prefix,
            feature_name_prefix=feature_name_prefix,
            enable_feature_lookup_output_to_csv=enable_feature_lookup_output_to_csv,
            enable_write_output=enable_write_output,
            return_all_afe_columns=return_all_afe_columns,
            enable_all_feature_lookup_output_to_csv=enable_all_feature_lookup_output_to_csv
        )

    @classmethod
    def get_settings(cls,
                     trx_descriptor: Union[TrxDescriptor, Dict],
                     target_descriptor: Union[TargetDescriptor, Dict],
                     target_record_source_list: List[Union[str, pd.DataFrame]],
                     afe_learning_type: Union[str, AfeLearningType],
                     output_settings: Union[AfeOutputSettings, Dict] = None,
                     algorithm_settings: AfeAlgorithmSettings = None,
                     auto_column_decider_settings: AutoColumnDeciderSettings = None,
                     number_of_cores: int = None,
                     max_number_of_transaction_samples: int = None,
                     max_number_of_target_samples: int = None,
                     **kwargs
                     ):
        """
        Generates AfeModellingSettings instance.

        :param trx_descriptor:
        :param target_descriptor:
        :param target_record_source_list:
        :param output_settings:
        :param afe_learning_type:
        :param algorithm_settings:
        :param auto_column_decider_settings:
        :param number_of_cores: Maximum number of cores to be used while modelling
        :param max_number_of_target_samples:
        :param max_number_of_transaction_samples:
        :return:
        """
        # pylint: disable=too-many-arguments, unused-argument
        return cls._get_afe_settings_reader().get_settings(
            trx_descriptor, target_descriptor, target_record_source_list, output_settings,
            None, None, afe_learning_type,
            algorithm_settings, auto_column_decider_settings, number_of_cores, max_number_of_transaction_samples,
            max_number_of_target_samples)

    @classmethod
    def persist_model_output(cls, model_output: AfeModelOutputType, location: str):
        """
        Saves model output to given location.

        :param model_output:
        :param location: file path
        """
        if not isinstance(location, str):
            raise ValueError("Location invalid. Enter a file path.")
        serialize_to_file(model_output, location, secret_key=AfeStaticObjects.get_secret_key())

    @classmethod
    def get_model(cls, location: str, secret_key: str = None, **kwargs) -> AfeModelOutputType:
        """
        Loads model output from given location

        :param location: file path.
        :param secret_key: secret key generated while saving model
        """
        # pylint: disable=unused-argument
        if not isinstance(location, str):
            raise NotImplementedError("Location invalid. Enter a file path.")
        secret_key = secret_key if secret_key is not None else AfeStaticObjects.get_secret_key()
        return deserialize_from_file(location, secret_key=secret_key)

    @staticmethod
    def get_process_peak_memory_usage():
        """Returns peak memory usage of process in bytes"""
        return process_info_helper.get_peak_memory_usage()

    @classmethod
    def create_afe_report(cls, model_output: AfeModelOutputType, use_lib="matplotlib"):
        """
        Shows afe report dashboard.

        :param model_output:
        :param use_lib: Visualization library to use. Valid values: "matplotlib", "plotly"
        """
        report_helper: BaseReportHelper = cls._get_report_helper(use_lib)
        report = report_helper.generate_afe_report(model_output)
        report_helper.generate_afe_report_dashboard(report)

    @classmethod
    def create_feature_report(cls, model_output: AfeModelOutputType, use_lib="matplotlib"):
        """
        Shows afe features dashboard.

        :param model_output:
        :param use_lib: Visualization library to use. Valid values: "matplotlib", "plotly"
        """
        report_helper: BaseReportHelper = cls._get_report_helper(use_lib)
        report = report_helper.generate_afe_report(model_output)
        report_helper.generate_afe_feature_dashboard(report)

    @classmethod
    def create_resource_usage_report(cls, model_output: AfeModelOutputType, use_lib="matplotlib"):
        """
        Shows resource usage dashboard.

        :param model_output:
        :param use_lib: Visualization library to use. Valid values: "matplotlib", "plotly"
        """
        report_helper: BaseReportHelper = cls._get_report_helper(use_lib)
        report = report_helper.generate_afe_report(model_output)
        report_helper.generate_resource_usage_dashboard(report)

    @classmethod
    def create_extended_feature_report(cls, model_output: AfeModelOutputType, use_lib="matplotlib"):
        """
        Shows extended feature report.

        :param model_output:
        :param use_lib: Visualization library to use. Valid values: "matplotlib", "plotly"
        """
        report_helper: BaseReportHelper = cls._get_report_helper(use_lib)
        report = report_helper.generate_afe_report(model_output)
        report_helper.generate_feature_extended_dashboard(report)

    @classmethod
    def create_feature_importance_report(cls, model_output: AfeModelOutputType, target_source_index: int = None,
                                         first_x_features: int = None, use_lib="matplotlib"):
        """
        Shows importance of afe features generated for given target.

        :param model_output:
        :param first_x_features: Number of columns to show in plot
        :param target_source_index: Index of target(In modelling target_record_source_list)
        :param use_lib: Visualization library to use. Valid values: "matplotlib", "plotly"
        """
        if model_output.afe_learning_type == AfeLearningType.Unsupervised:
            raise KnownException("Feature importance report is not available on Unsupervised models.")
        report_helper: BaseReportHelper = cls._get_report_helper(use_lib)
        report = report_helper.generate_afe_report(model_output)
        report_helper.generate_feature_importance_dashboard(report, target_index=target_source_index,
                                                            first_x_feature=first_x_features)

    @staticmethod
    def generate_feature_importance_report_for_targets(model_output: AfeModelOutputType) -> List[pd.DataFrame]:
        """
        Returns importance of afe features generated for all targets.

        :param model_output:
        :return: List of pandas dataframes storing feature importance data for targets
        """
        if model_output.afe_learning_type == AfeLearningType.Unsupervised:
            raise KnownException("Feature importance report is not available on Unsupervised models.")
        importance_data_list = []
        if model_output.final_column_metrics is None:
            raise ValueError("afe report does not include final column metrics")
        for metrics in model_output.final_column_metrics:
            data = metrics["feature_importances"].sort_values(by="Value", ascending=False)
            importance_data_list.append(data)
        return importance_data_list

    @staticmethod
    def generate_auc_score_for_targets(model_output: AfeModelOutputType, target_index: int = None) -> List[
        Tuple[int, float]]:
        """
         Returns auc score for all targets.
        :param model_output:
        :return: List of float storing auc score for targets

        """
        if model_output.afe_learning_type == AfeLearningType.Unsupervised:
            raise KnownException("AUC score is not available on Unsupervised models.")
        if model_output.final_column_metrics is None:
            raise ValueError("afe report does not include final column metrics")
        metrics_to_report = list(enumerate(model_output.final_column_metrics))
        auc_score_list = []

        if target_index is not None:
            if len(model_output.final_column_metrics) <= target_index:
                raise KnownException(f"Target with index {target_index} does not exist. "
                                     f"Num targets: {len(model_output.final_column_metrics)}")
            auc_score_list.append((target_index, model_output.final_column_metrics[target_index]["auc"]["all"]))
        else:
            for i, metrics in metrics_to_report:
                auc = metrics["auc"]["all"]
                auc_score_list.append((i, auc))
        return auc_score_list

    @classmethod
    def create_roc_curve_report(cls, model_output: AfeModelOutputType, target_source_index: int = None,
                                target_class=None, use_lib="matplotlib"):
        """
        Shows roc_curve for given target.

        :param model_output:
        :param target_source_index: Index of target(In modelling target_record_source_list)
        :param target_class: target class report roc_curve for (if multiclass afe modelling)
        :param use_lib: Visualization library to use. Valid values: "matplotlib", "plotly"
        """
        if model_output.afe_learning_type == AfeLearningType.Unsupervised:
            raise KnownException("ROC report is not available on Unsupervised models.")
        report_helper: BaseReportHelper = cls._get_report_helper(use_lib)
        report = report_helper.generate_afe_report(model_output)
        report_helper.generate_roc_curve_dashboard(report, target_index=target_source_index, target_class=target_class)

    @classmethod
    def create_features_data_frame(cls, model_output: AfeModelOutputType):
        """
        Returns afe features as dataframe.

        :param model_output:
        :return:
        """
        return PersistHelper.get_lookup_data_frame(list(model_output.output_features.values()),
                                                   model_output.transaction_file_stats)

    @classmethod
    def create_all_features_data_frame(cls, model_output: AfeModelOutputType):
        """
        Returns all generated afe features as dataframe.

        :param model_output:
        :return:
        """
        return PersistHelper.get_lookup_data_frame(list(model_output.all_features.values()),
                                                   model_output.transaction_file_stats)

    @classmethod
    def _get_report_helper(cls, use_lib="matplotlib") -> BaseReportHelper:
        return AfeMatplotlibReportHelper() if use_lib == "matplotlib" else AfePlotlyReportHelper()

    @staticmethod
    def set_secret_key(secret_key: str):
        """Sets secret key for encryption tasks used during serialization etc."""

        AfeStaticObjects.set_secret_key(secret_key)



"""Includes helper functions for ensembling"""
from typing import List, Union

import numpy as np
import pandas as pd

from organon.ml.modelling.algorithms.core.enums.modeller_type import ModellerType


def get_voting_prediction(all_predictions: Union[pd.DataFrame, np.ndarray], modeller_type: ModellerType) -> pd.Series:
    """Returns prediction after applying voting on 'all_predictions' which stores different predictions as columns"""
    if modeller_type == ModellerType.REGRESSOR:
        return get_voting_regressor_prediction(all_predictions)
    if modeller_type == ModellerType.CLASSIFIER:
        return get_voting_classifier_prediction(all_predictions)
    raise NotImplementedError


def get_voting_classifier_prediction(all_predictions: Union[pd.DataFrame, np.ndarray]) -> pd.Series:
    """Returns prediction after applying voting on 'all_predictions' which stores different predictions as columns"""
    final_pred = np.apply_along_axis(_get_most_observed_val, axis=1, arr=all_predictions)
    if final_pred.ndim > 1:
        final_pred = final_pred[:, 0]
    return pd.Series(final_pred)


def get_voting_regressor_prediction(all_predictions: Union[pd.DataFrame, np.ndarray]) -> pd.Series:
    """Returns prediction after applying voting on 'all_predictions' which stores different predictions as columns"""
    final_pred = np.mean(all_predictions, axis=1)
    return pd.Series(final_pred)


def get_voting_predict_proba(proba_dfs: List[pd.DataFrame]):
    """Returns probabilities after applying voting on 'proba_dfs' where every dataframe is calculated
    by a different modeller"""
    columns = proba_dfs[0].columns.tolist()
    proba_dfs = [proba_df.loc[:, columns] for proba_df in proba_dfs]
    average_probas = sum(proba_dfs) / len(proba_dfs)
    return average_probas


def _get_most_observed_val(arr: np.array):
    """returns most observed value in array. If there are two with same number of occurrences, chooses randomly"""
    unique, counts = np.unique(arr, return_counts=True)
    indices = np.where(counts == max(counts))
    if len(indices[0]) > 1:
        return np.random.choice(unique[indices], 1)
    return unique[indices][0]



"""
This module keeps the functions that are used in Framework Library initialization.
"""
import threading

from organon.fl.core.businessobjects.fl_core_static_objects import FlCoreStaticObjects

FL_INIT_LOCK = threading.Lock()


class FlInitializer:
    """Class for fl application operations"""

    @classmethod
    def application_initialize(cls):
        """
        Gives the signal for application initialization the application by passing True value to the
        application_initialize_called variable.
        :return: nothing
        """
        with FL_INIT_LOCK:
            if not FlCoreStaticObjects.application_initialize_called:
                cls._on_init()
                FlCoreStaticObjects.application_initialize_called = True

    @classmethod
    def _on_init(cls):
        cls.register_types()

    @classmethod
    def register_types(cls):
        """
        Registers the IoC items according to its ioc key, abstract type and concrete type.
        :return:nothing
        """



"""Includes CFS class."""
from typing import Union, Optional

import pandas as pd

from organon.ml.common.enums.target_type import TargetType
from organon.ml.common.helpers.user_input_service_helper import get_enum
from organon.ml.feature_selection.domain.objects.cfs_selection_output import CFSSelectionOutput
from organon.ml.feature_selection.domain.objects.settings.cfs_selection_settings import CFSSelectionSettings
from organon.ml.feature_selection.domain.services.cfs_selection_service import CFSSelectionService
from organon.ml.feature_selection.services.abstractions.base_supervised_feature_selecter import \
    BaseSupervisedFeatureSelecter


class CFS(BaseSupervisedFeatureSelecter):
    """Feature selecter which utilizes a correlation based feature selection algorithm"""

    def __init__(self, data: pd.DataFrame, target: Union[pd.DataFrame, pd.Series], target_type:str, *,
                 max_backtracks: int = 5, prioritize_performance=False):
        super().__init__(data, target)
        target = self._get_target_as_df(target)
        self.settings = CFSSelectionSettings(data, target, get_enum(target_type, TargetType),
                                             max_backtracks=max_backtracks,
                                             prioritize_performance=prioritize_performance)
        self.output: Optional[CFSSelectionOutput] = None

    def _run(self):
        service = CFSSelectionService()
        self.output = service.run_selection(self.settings)
        return self.output



"""todo"""
from typing import List

from organon.fl.core.enums.column_native_type import ColumnNativeType


class DqDataColumn:
    """todo"""

    def __init__(self):
        self.column_name: str = None
        self.column_description: str = None
        self.default_values: List[str] = None
        self.column_native_type: ColumnNativeType = None
        self.default_values_numeric: List[float] = None
        self.default_values_nominal: List[str] = None
        self.inclusion_flag: bool = None

    # def __initialize(self):
    #     self.column_name = ""
    #     self.column_description = ""
    #     self.data_type = object
    #     self.column_native_type = ColumnNativeType.Other
    #     self.default_values = []



"""Includes UserExistingCalculationParams"""
from organon.idq.domain.businessobjects.dq_calculation_output import DqCalculationOutput


class UserExistingCalculationParams:
    """Stores existing calculation parameters entered by user"""
    def __init__(self, calculation_output: DqCalculationOutput, calculation_name: str):
        self.calculation_output: DqCalculationOutput = calculation_output
        self.calculation_name = calculation_name



"""
This module includes Transforamtion class.
"""
import locale
from typing import List, Dict
from sortedcontainers import SortedDict

from organon.fl.core.helpers import list_helper


class Transformation:
    """AFE model output transformation"""
    def __init__(self, default_input_list: List[float], default_output: float, map_: SortedDict):
        self.__default_input_list: List[float] = default_input_list
        self.__default_output: float = default_output
        self.__map: SortedDict = map_
        self.__default_input_list.sort()

    @property
    def default_input_list(self):
        """Returns value of private attribute __default_input_list"""
        return self.__default_input_list

    @property
    def default_output(self):
        """Returns value of private attribute __default_output"""
        return self.__default_output

    @property
    def map_(self):
        """Returns value of private attribute __map"""
        return self.__map

    def transform(self, value_to_map: float) -> float:
        """
        Finds corresponding value according to transformation map
        :param value_to_map: map
        :return: Mapped value
        """
        if list_helper.binary_search(self.__default_input_list, value_to_map) != -1:
            return self.__default_output

        for key in self.__map:
            if value_to_map <= key:
                return self.__map[key]
        raise NotImplementedError

    def transform_all(self, input_: List[float]):
        """Transforms all values in list(in place)."""
        for i, val in enumerate(input_):
            input_[i] = self.transform(val)

    def tabulate_as_text(self) -> Dict[str, str]:
        """Converts mappings to a dictionary where keys are ranges of mappings and values are transform values"""
        result: Dict[str, str] = {}
        rounded_default_output = locale.str(round(self.__default_output, 6))
        for _input in self.__default_input_list:
            rounded_value = locale.str(round(_input, 6))
            result[rounded_value] = rounded_default_output
        if self.__map is None or len(self.__map) == 0:
            return result

        keys = list(self.__map.keys())
        values = list(self.__map.values())

        for i, key in enumerate(keys):
            down_string = "-Infty" if i == 0 else locale.str(round(keys[i - 1], 6))
            up_string = "+Infty" if i == len(keys)-1 else locale.str(round(key, 6))
            interval = f"[{down_string}, {up_string})"
            result[interval] = locale.str(round(values[i], 6))

        return result



"""Includes LogisticRegressionClassifier class."""
from sklearn.linear_model import LogisticRegression

from organon.ml.modelling.algorithms.core.abstractions.base_sklearn_classifier import BaseSklearnClassifier


class LogisticRegressionClassifier(BaseSklearnClassifier[LogisticRegression]):
    """Wrapper for sklearn LogisticRegression"""

    @property
    def classifier_class(self):
        return LogisticRegression



"""Includes DqBaseCalculationParameters class."""
from typing import TypeVar, Generic, List

from organon.fl.core.enums.column_native_type import ColumnNativeType
from organon.idq.domain.businessobjects.dq_calculation_result import DqCalculationResult
from organon.idq.domain.settings.abstractions.dq_base_input_source_settings import DqBaseInputSourceSettings
from organon.idq.domain.settings.dq_column_metadata import DqColumnMetadata

InputSourceSettingsType = TypeVar("InputSourceSettingsType", bound=DqBaseInputSourceSettings)


class DqBaseCalculationParameters(Generic[InputSourceSettingsType]):
    """Base calculation settings for all source types"""

    def __init__(self):
        self.calculation_name: str = None
        self.input_source_settings: InputSourceSettingsType = None
        self.max_nominal_cardinality_count: int = None
        self.outlier_parameter: float = None
        self.column_dq_metadata_list: List[DqColumnMetadata] = None
        self.use_population_nominal_stats: bool = None
        self.is_existing_calculation: bool = None
        self.existing_result: DqCalculationResult = None
        self.nominal_column_types: List[ColumnNativeType] = None



"""Includes UserInputService class."""
from typing import List, Dict

import pandas as pd

from organon.fl.core.enums.column_native_type import ColumnNativeType
from organon.idq.domain.enums.dq_record_source_type import DqRecordSourceType
from organon.idq.domain.services.base_user_input_service import BaseUserInputService
from organon.idq.domain.services.user_input_validation_service import UserInputValidationService
from organon.idq.services.user_settings.dq_user_input import DqUserInput


class UserInputService(BaseUserInputService[DqUserInput]):
    """Includes methods to read and validate dq user input"""

    def __init__(self, user_input: DqUserInput):
        super().__init__(user_input)

    def _get_input_validation_service(self, source_columns: List[Dict[str, ColumnNativeType]]) \
            -> UserInputValidationService:
        return UserInputValidationService(self._user_input, source_columns[0])

    def _get_columns_in_sources(self) -> List[Dict[str, ColumnNativeType]]:
        self._validate_input_source_settings(self._user_input.calc_params.input_source_settings)
        return [self._get_column_native_types(self._user_input.calc_params.input_source_settings)]

    def _get_source_type(self):
        comp_source = self._user_input.calc_params.input_source_settings.source
        if isinstance(comp_source, str):
            return DqRecordSourceType.TEXT
        if isinstance(comp_source, pd.DataFrame):
            return DqRecordSourceType.DATA_FRAME
        raise ValueError("Invalid input source type!")

    def _get_calc_params(self, source_columns: List[Dict[str, ColumnNativeType]]):
        return self._get_calculation_params_from_user_multi_part_calc_params(self._user_input.calc_params,
                                                                             list(source_columns[0].keys()))



"""Includes BlasServiceFactory class."""
from organon.fl.mathematics.linearalgebra.IBlas import IBlas
from organon.fl.mathematics.linearalgebra.MklFunctions import MklFunctions, MKL_FOUND
from organon.fl.mathematics.linearalgebra.blas_service_type import BlasServiceType
from organon.fl.mathematics.linearalgebra.scipy_blas_service import ScipyBlasService


class BlasServiceFactory:
    """Factory for IBlas implementations."""

    @staticmethod
    def get_blas_service(blas_type: BlasServiceType = None) -> IBlas:
        """Returns a service with IBlas implementations according to given blas_type."""
        if blas_type is None:
            if MKL_FOUND:
                return MklFunctions()
            return ScipyBlasService()
        if blas_type == BlasServiceType.MKL_BLAS:
            return MklFunctions()
        if blas_type == BlasServiceType.SCIPY_BLAS:
            return ScipyBlasService()
        raise NotImplementedError



"""Includes RFSelectionSettings class"""
from typing import Any, Dict, Optional

import pandas as pd

from organon.ml.common.enums.target_type import TargetType
from organon.ml.feature_selection.domain.objects.settings.base_supervised_feature_selection_settings import \
    BaseSupervisedFeatureSelectionSettings


class RFSelectionSettings(BaseSupervisedFeatureSelectionSettings):
    """Settings for random forest selection"""

    def __init__(self, data: pd.DataFrame, target: pd.DataFrame, target_type: TargetType = None,
                 num_features: Optional[int] = None,
                 sweep_args: dict = None,
                 rf_args: Dict[str, Any] = None):
        super().__init__(data, target, target_type)
        self.rf_args = rf_args
        self.num_features = num_features
        self.sweep_args = sweep_args



""""This module includes ModellingReductionSetttings class."""

from organon.afe.domain.modelling.businessobjects.target_file_record_collection import TargetFileRecordCollection
from organon.afe.domain.modelling.businessobjects.transaction_file_record_collection import \
    TransactionFileRecordCollection
from organon.afe.domain.settings.record_source import RecordSource


class ModellingReductionSettings:
    """Settings for afe column elimination in modelling"""

    def __init__(self):
        self.target_record_source: RecordSource = None
        self.target_record_source_index: int = None
        self.target_file_record_collection: TargetFileRecordCollection = None
        self.trx_file_record_collection: TransactionFileRecordCollection = None
        self.num_threads: int = None



"""Includes UserSelectionSettings class"""
from dataclasses import dataclass
from typing import List

import pandas as pd

from organon.ml.modelling.algorithms.core.abstractions.base_modeller import BaseModeller


@dataclass
class UserSelectionSettings:
    """Settings entered by user for Selecter"""
    modellers: List[BaseModeller]
    train_data: pd.DataFrame
    target_data: pd.Series
    target_type: str
    cv_count: int
    test_data: pd.DataFrame
    test_target_data: pd.Series
    add_stacking: bool
    add_voting: bool
    num_threads_for_parallel_fit: int = None
    num_threads_for_parallel_cv: int = None



"""Includes LGBMClassifier class."""
from lightgbm import LGBMClassifier

from organon.ml.modelling.algorithms.core.abstractions.base_sklearn_classifier import BaseSklearnClassifier


class LightGBMClassifier(BaseSklearnClassifier[LGBMClassifier]):
    """Wrapper for LGBMClassifier"""

    @property
    def classifier_class(self):
        return LGBMClassifier



"""Includes BaseSklearnRegressor class."""
import abc
from typing import TypeVar, Generic, Type, Dict, Any, Union

import pandas as pd

from organon.ml.modelling.algorithms.core.abstractions.base_regressor import BaseRegressor

SklearnRegressorType = TypeVar("SklearnRegressorType")


class BaseSklearnRegressor(Generic[SklearnRegressorType], BaseRegressor, metaclass=abc.ABCMeta):
    """Base class for sklearn classifiers"""

    def __initialize_with_params(self, **params):
        # pylint: disable=attribute-defined-outside-init
        self._modeller: SklearnRegressorType = self.regressor_class(**params)

    @property
    @abc.abstractmethod
    def regressor_class(self) -> Type[SklearnRegressorType]:
        """Returns wrapped sklearn regressor class"""

    def _fit(self, train_data: pd.DataFrame, target_data: Union[pd.DataFrame, pd.Series], **kwargs):
        if isinstance(target_data, pd.DataFrame):
            target_data = target_data[target_data.columns[0]]
        return self._modeller.fit(train_data, target_data, **kwargs)

    def _predict(self, data: pd.DataFrame) -> pd.DataFrame:
        return pd.DataFrame(self._modeller.predict(data))

    def _score(self, train_data: pd.DataFrame, target_data: Union[pd.DataFrame, pd.Series],
               sample_weight=None) -> float:
        return self._modeller.score(train_data, target_data)

    def _get_params_with_defaults(self) -> Dict[str, Any]:
        return self.regressor_class().get_params()

    def get_params(self, deep=True):
        return self._modeller.get_params(deep=deep)

    def set_params(self, **params):
        self.__initialize_with_params(**params)
        return self



"""
This module include DbObjectInput class.
"""
from organon.afe.domain.common.reader_helper import get_values_from_kwargs

DbObjClass = object
try:
    from organon.fl.dataaccess.datadtos.db_object_dto import DbObjectDto

    DbObjClass = DbObjectDto  # pycharm warninglerin kaldırılması için böyle yapıldı
except ImportError:
    pass


class DbObjectInput(DbObjClass):  # pylint: disable=useless-object-inheritance
    """Db Object with connection name specified"""

    ATTR_DICT = {
        "schema_name": str,
        "table_name": str,
        "sql_statement": str,
        "connection_name": str
    }

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.connection_name = None

        get_values_from_kwargs(self, self.ATTR_DICT, kwargs)



"""This module includes PersistHelper class."""
from typing import List, Dict

import pandas as pd

from organon.afe.core.businessobjects.afe_static_objects import AfeStaticObjects
from organon.afe.domain.enums.afe_operator import AfeOperator
from organon.afe.domain.enums.date_resolution import DateResolution
from organon.afe.domain.modelling.businessobjects.afe_column import AfeColumn
from organon.afe.domain.modelling.businessobjects.base_afe_feature import BaseAfeFeature
from organon.afe.domain.modelling.businessobjects.transaction_file_stats import TransactionFileStats
from organon.fl.core.exceptionhandling.known_exception import KnownException
from organon.fl.core.helpers import string_helper


class PersistHelper:
    """Includes helper methods for persisting AFE results in db or file"""

    @classmethod
    def get_lookup_data_frame(cls, features: List[BaseAfeFeature],
                              stats: TransactionFileStats) -> pd.DataFrame:
        """Generates afe lookup data table"""
        if len(features) == 0:
            raise KnownException("Output features are empty.")

        collection_len = len(features)
        rows_per_step = max(1000, int(collection_len / 15))
        counter = 1
        rows_to_append = []
        data_frame: pd.DataFrame = None
        for item in features:
            row = cls._get_row_dict_from_feature(item, stats)
            rows_to_append.append(row)
            if counter % rows_per_step == 0 or counter == collection_len:
                if data_frame is None:
                    data_frame = pd.DataFrame(rows_to_append)
                else:
                    data_frame = pd.concat([data_frame, pd.DataFrame(rows_to_append)])
                rows_to_append.clear()
            counter += 1
        data_frame.reset_index(drop=True, inplace=True)
        return data_frame

    @classmethod
    def _get_row_dict_from_feature(cls, item: BaseAfeFeature, stats: TransactionFileStats):
        afe_column = item.afe_column
        feature_name = item.feature_name
        dimension_name = afe_column.dimension_name
        histogram = stats.get_histogram(dimension_name)
        _map = histogram.reverse_index
        feature_extended_name = PersistHelper.__get_feature_extended_name(afe_column, histogram)
        dimension_set = ""
        if afe_column.operator not in [AfeOperator.Mode, AfeOperator.CountDistinct]:
            dimension_set = afe_column.get_dimension_set(_map)
        row = {
            "FEATURE": feature_name,
            "FEATURE_EXTENDED_NAME": feature_extended_name,
            "SOURCE": afe_column.source,
            "DIMENSION": dimension_name,
            "DIMENSION_SET": dimension_set,
            "QUANTITY": afe_column.quantity_name,
            "OPERATOR": afe_column.operator.name,
            "TIME_WINDOW": afe_column.time_window,
            "RESOLUTION": afe_column.date_resolution.name,
            "DATE_OFFSET": afe_column.offset,
            "MEANING": cls.get_feature_meaning(item, _map)
        }
        if afe_column.date_column is not None:
            row["DATE_COLUMN_NAME"] = afe_column.date_column.column_name
        else:
            row["DATE_COLUMN_NAME"] = None
        return row

    @staticmethod
    def __get_feature_extended_name(afe_column, histogram):
        feature_extended_name = afe_column.build_column_name_with_map(histogram.reverse_index)
        if len(feature_extended_name) >= 100:
            feature_extended_name = feature_extended_name[:97] + "..."
        return feature_extended_name

    @staticmethod
    def print_report(selected_cols, file, d_col: str, q_col: str, name_to_column: Dict[str, AfeColumn],
                     stats: TransactionFileStats):
        """Prints model variables to file"""

        if not string_helper.is_null_or_empty(d_col) and not string_helper.is_null_or_empty(q_col):
            file.write(f"Non-interaction model for dimension column: {d_col} quantity column-{q_col}:\n")

        # file.write(f"Performance. R2:{}, AUC:{}\n")

        for column_name in selected_cols:
            afe_column = name_to_column[column_name]
            histogram = stats.get_histogram(afe_column.dimension_name)
            modified_name = afe_column.build_column_name_with_map(histogram.reverse_index)
            file.write(f"{modified_name}\n")

        file.write("\n**********************************\n")
        file.flush()

    @classmethod
    def get_feature_meaning(cls, feature: BaseAfeFeature, group_names: Dict[int, str]) -> str:
        """Returns meaning of given afe output feature"""
        afe_column = feature.afe_column
        is_not_case = not afe_column.in_out
        is_empty_case = len(afe_column.set) == 0
        qty_check = afe_column.quantity_name == AfeStaticObjects.empty_quantity_column
        dim_check = afe_column.dimension_name == AfeStaticObjects.empty_dimension_column

        operator = afe_column.operator

        before_last_dict = cls._get_before_last_dict(afe_column)

        resolution_dict = {
            DateResolution.Year: "years",
            DateResolution.Month: "months",
            DateResolution.Day: "days",
            DateResolution.Hour: "hours",
            DateResolution.Minute: "minutes",
            DateResolution.Second: "seconds"
        }

        dimension_str = afe_column.get_dimension_set(group_names)
        dimension_desc = dimension_str
        if is_not_case:
            dimension_desc = dimension_str[1:]

        where_d = ""

        if not dim_check:
            where_d = f"{afe_column.dimension_name} is {dimension_desc}"
            if is_not_case:
                where_d = f"{afe_column.dimension_name} is not {dimension_desc}"
            if not is_not_case and is_empty_case:
                where_d = f"{afe_column.dimension_name} is NULL"

        where_condition = where_d
        if not qty_check:
            if not string_helper.is_null_or_empty(where_condition):
                where_condition += " and "
            where_quantity_dict = cls._get_where_quantity_str_dict(afe_column)
            where_q = where_quantity_dict[operator]
            where_condition += where_q

        where = "" if qty_check and dim_check else "where"

        time_window_dict = cls._get_time_window_str_dict(afe_column)
        days = time_window_dict[operator]
        resolution_str = resolution_dict[DateResolution[afe_column.date_resolution.name]]

        if operator == AfeOperator.Mode:
            return f"most observed {afe_column.dimension_name} value in last {days} {resolution_str}"
        if operator == AfeOperator.CountDistinct:
            return f"Number of distinct {afe_column.dimension_name} values in last {days} {resolution_str}"
        before_last = before_last_dict[operator]
        return f"{before_last} last {days} {resolution_str} {where} {where_condition}"

    @classmethod
    def _get_before_last_dict(cls, afe_column: AfeColumn) -> Dict[AfeOperator, str]:
        quantity = afe_column.quantity_name
        if quantity and quantity != AfeStaticObjects.empty_quantity_column:
            return {
                AfeOperator.Min: f"minimum {quantity} in",
                AfeOperator.Max: f"maximum {quantity} in",
                AfeOperator.Sum: f"{quantity} sum in",
                AfeOperator.Frequency: "count in",
                AfeOperator.Ratio: f"{quantity} sum ratio in",
                AfeOperator.Density: f"{quantity} average of",
                AfeOperator.TimeSinceFirst: "time since first record in",
                AfeOperator.TimeSinceLast: "time since last record in"
            }
        return {
            AfeOperator.Min: "record count ratio of",
            AfeOperator.Max: "record count ratio of",
            AfeOperator.Sum: "record count in",
            AfeOperator.Frequency: "record count in",
            AfeOperator.Ratio: "record count ratio in",
            AfeOperator.Density: "record existing info in",
            AfeOperator.TimeSinceFirst: "time since first record in",
            AfeOperator.TimeSinceLast: "time since last record in",
        }

    @classmethod
    def _get_time_window_str_dict(cls, afe_column: AfeColumn) -> Dict[AfeOperator, str]:
        timewindow = str(afe_column.time_window)
        return {
            AfeOperator.Min: timewindow,
            AfeOperator.Max: timewindow,
            AfeOperator.Sum: timewindow,
            AfeOperator.Frequency: timewindow,
            AfeOperator.Ratio: timewindow,
            AfeOperator.Density: timewindow,
            AfeOperator.TimeSinceFirst: timewindow,
            AfeOperator.TimeSinceLast: timewindow,
            AfeOperator.Mode: timewindow,
            AfeOperator.CountDistinct: timewindow
        }

    @classmethod
    def _get_where_quantity_str_dict(cls, afe_column: AfeColumn) -> Dict[AfeOperator, str]:
        quantity = afe_column.quantity_name
        return {
            AfeOperator.Min: f"{quantity} >0",
            AfeOperator.Max: f"{quantity} >0",
            AfeOperator.Sum: f"{quantity} >0",
            AfeOperator.Frequency: f"{quantity} >0",
            AfeOperator.Ratio: f"{quantity} > 0 to where {quantity} > 0",
            AfeOperator.Density: f"{quantity} >0",
            AfeOperator.TimeSinceFirst: f"{quantity} >0",
            AfeOperator.TimeSinceLast: f"{quantity} >0"
        }



"""Includes RFSelectionService class"""
import pandas as pd
from sklearn.ensemble import RandomForestClassifier, RandomForestRegressor

from organon.ml.common.enums.target_type import TargetType
from organon.ml.feature_selection.domain.objects.rf_selection_output import RFSelectionOutput
from organon.ml.feature_selection.domain.objects.settings.rf_selection_settings import RFSelectionSettings
from organon.ml.feature_selection.domain.services.base_feature_selection_service import SelectionRunSettings
from organon.ml.feature_selection.domain.services.base_supervised_feature_selection_service import \
    BaseSupervisedFeatureSelectionService


class RFSelectionService(BaseSupervisedFeatureSelectionService[RFSelectionSettings, RFSelectionOutput]):
    """Service class for RFSelection"""

    @classmethod
    def _validate_settings(cls, settings: RFSelectionSettings):
        cls._validate_original_data_for_sweep(settings.data)

    @classmethod
    def _run(cls, run_settings: SelectionRunSettings) -> RFSelectionOutput:
        settings: RFSelectionSettings = run_settings.settings
        data = run_settings.data
        estimator_args = {} if settings.rf_args is None else settings.rf_args
        rf_estimator = RandomForestClassifier(**estimator_args) \
            if settings.target_type in [TargetType.BINARY, TargetType.MULTICLASS] \
            else RandomForestRegressor(**estimator_args)
        rf_estimator.fit(data, settings.target)
        imp_df = pd.DataFrame(rf_estimator.feature_importances_, columns=['importance'], index=data.columns)
        imp_df = imp_df.sort_values(by="importance", ascending=False)
        num_features = settings.num_features
        if num_features is None:
            num_features = len(cls._get_features_with_sweep(data, settings.sweep_args))
        selected_features = imp_df.index[:num_features].to_list()
        output = RFSelectionOutput()
        output.selected_features = selected_features
        return output



"""Module for CscMatrix"""
import numpy as np
from organon.fl.mathematics.linearalgebra.MklFunctions import MklFunctions
from organon.fl.mathematics.linearalgebra.enums import descr, SparseOperationT
from organon.fl.mathematics.sweep.matrices.i_matrix import IMatrix


class CscMatrix(IMatrix):
    """Double Csr matrix for sparse matrices"""
    @property
    def data(self) -> np.ndarray:
        """Get the data"""
        return self._data

    @property
    def col_ptr(self) -> np.ndarray:
        """Get col ptr"""
        return self._col_ptr

    @property
    def row_ptr(self) -> np.ndarray:
        """Get the row ptr"""
        return self._row_ptr

    @property
    def shape(self) -> tuple:
        """Get the shape of matrix"""
        return self._shape

    @property
    def means(self) -> np.ndarray:
        """Get column means of the matrix"""
        return self._means

    @property
    def is_centralized(self) -> bool:
        """Get the centralized information"""
        return self._is_centralized

    def __init__(self, data: np.ndarray, row_ptr: np.ndarray, col_ptr: np.ndarray, shape: tuple = None):

        CscMatrix.validate(data, row_ptr, col_ptr, shape)

        self._nnz = data.size
        self._data = data
        self._row_ptr = row_ptr
        self._col_ptr = col_ptr

        if shape is not None:
            self._shape = shape
        else:
            self._shape = (max(row_ptr) + 1, col_ptr.size - 1)

        self._means = np.empty(self._shape[1], dtype=data.dtype)
        self._calculate_means()

        self._is_centralized = False

    def _calculate_means(self):
        for i in range(self._shape[1]):
            start = self._col_ptr[i]
            end = self._col_ptr[i + 1]
            self._means[i] = np.sum(self._data[start:end])
        self._means /= self._shape[0]

    @staticmethod
    def validate(data: np.ndarray, row_ptr: np.ndarray, col_ptr: np.ndarray, shape: tuple):
        """Validates the parameters"""
        if data is None or row_ptr is None or col_ptr is None:
            raise ValueError("Arrays should be defined")

        if data.dtype.type not in (np.single, np.double):
            raise ValueError("Data Type must be single or double")

        if row_ptr.size != data.size:
            raise ValueError("data size and col_ptr size are incompatible")

        if col_ptr[-1] != data.size:
            raise ValueError("col_ptr[-1] and NNZ are incompatible")

        if shape is not None:
            if len(shape) != 2:
                raise ValueError("Array must be 2 dimensional")
            if shape[1] != col_ptr.size - 1:
                raise ValueError("Column count and col_ptr size are incompatible")

            if shape[0] < max(row_ptr) + 1:
                raise ValueError("Row count and row_ptr elements are incompatible")

    def get_column(self, k: int) -> np.ndarray:
        """Get specified column of the matrix"""
        column_array = np.zeros(self._shape[0], dtype=self._data.dtype)

        start = self._col_ptr[k]
        end = self._col_ptr[k + 1]

        column_array[self._row_ptr[start:end]] = self._data[start:end]

        return column_array

    def centralize_columns(self):
        """Centralize all columns"""

    def to_dense(self) -> np.ndarray:
        """Convert sparse matrix to dense matrix"""
        dense_matrix = np.zeros(shape=self._shape, dtype=self._data.dtype)

        for j in range(len(self._col_ptr) - 1):
            start = self._col_ptr[j]
            end = self._col_ptr[j + 1]

            for i in range(start, end):
                dense_matrix[self._row_ptr[i], j] = self._data[i]

        return dense_matrix

    def mul_vector(self, vector: np.ndarray) -> np.ndarray:
        """Calculates Matrix Vector product"""
        dtype = self._data.dtype
        result = np.zeros(shape=self._shape[1], dtype=dtype)
        if dtype.type is np.single:
            MklFunctions.sparse_s_mv(SparseOperationT.SPARSE_OPERATION_TRANSPOSE, self._shape[0], self._shape[1],
                                     np.single(1.0),
                                     descr, self._data, self._row_ptr,
                                     self._col_ptr[0:-1],
                                     self._col_ptr[1:], vector, np.single(1.0), result, sparse_format='csc')
        else:
            MklFunctions.sparse_d_mv(SparseOperationT.SPARSE_OPERATION_TRANSPOSE, self._shape[0], self._shape[1],
                                     np.double(1.0),
                                     descr, self._data, self._row_ptr,
                                     self._col_ptr[0:-1],
                                     self._col_ptr[1:], vector, np.double(1.0), result, sparse_format='csc')
        return result



"""Includes FeatureSimilaritySelectionSettings class"""
import pandas as pd

from organon.ml.feature_selection.domain.objects.settings.base_unsupervised_feature_selection_settings import \
    BaseUnsupervisedFeatureSelectionSettings


class FeatureSimilaritySelectionSettings(BaseUnsupervisedFeatureSelectionSettings):
    """Settings for feature similarity selection"""

    def __init__(self, data: pd.DataFrame, selection_percent: float = 0.05, prioritize_performance: bool = False):
        super().__init__(data)
        self.selection_percent: float = selection_percent
        self.prioritize_performance = prioritize_performance



"""
This module includes methods to serialize objects.
"""
from enum import Enum
from typing import Dict

import numpy as np

from organon.fl.core.helpers import object_helper


def serialize_to_file(obj, filepath: str, secret_key: str = None):
    """
    Binary serialization of object to given file

    :returns : secret key used in encryption of pickled data. should be saved for deserialization
    """

    with open(filepath, "wb") as file:
        byte_val, secret_key = object_helper.get_encrypted_binary(obj, secret_key=secret_key)
        file.write(byte_val)
    return secret_key


def deserialize_from_file(filepath: str, secret_key: str):
    """Binary deserialization object from file."""
    with open(filepath, "rb") as readfile:
        obj = object_helper.get_object_from_encrypted_binary(readfile.read(), secret_key)
    return obj


def normalize_dictionary_for_json_conversion(_d: Dict):
    """Normalizes dictionary keys to become usable with json.dump"""
    type_map = {
        float: [float, np.float, np.float32],
        int: [int, np.int, np.int16, np.int32]
    }
    new_d = {}
    for key in _d:
        val = _d[key]
        val_type = type(val)

        if isinstance(val, dict):
            val = normalize_dictionary_for_json_conversion(val)
        elif val_type in type_map[float]:
            val = float(val)
        if isinstance(key, Enum):
            new_d[key.name] = val
        else:
            new_d[str(key)] = val

    return new_d



"""Includes ScalingService class"""
from typing import List

import pandas as pd
from sklearn.preprocessing import MinMaxScaler
from sklearn.preprocessing import StandardScaler

from organon.fl.core.enums.column_native_type import ColumnNativeType
from organon.fl.core.helpers.data_frame_helper import get_column_native_type
from organon.ml.preprocessing.settings.enums.scaler_type import ScalerType
from organon.ml.preprocessing.settings.objects.scaling_settings import ScalingSettings


class ScalingService:
    """Scales numeric data according to scaling strategy"""

    def __init__(self, settings: ScalingSettings):
        self._settings = settings
        self._scaler_per_col: dict = None
        self._numerical_cols: List[str] = None

    def columns_(self):
        """Returns fitted columns to scale"""
        return self._numerical_cols

    def fit(self, data: pd.DataFrame):
        """Fits scaler"""
        self._scaler_per_col = {}
        self._numerical_cols = [col for col in data.columns if
                                get_column_native_type(data, col) == ColumnNativeType.Numeric]
        for col in self._numerical_cols:
            scaler = self._get_scaler(self._settings.strategy)
            scaler.fit(data[col].values.reshape(-1, 1))
            self._scaler_per_col[col] = scaler
        return self

    def transform(self, data: pd.DataFrame, inplace: bool = False):
        """Transform data"""
        if self._scaler_per_col is None:
            raise ValueError("Scaler not fitted yet")
        if not inplace:
            trans_data = data.copy()
        else:
            trans_data = data
        different_columns = [col for col in trans_data if
                             (get_column_native_type(trans_data, col) == ColumnNativeType.Numeric and
                              col not in self._numerical_cols)]
        if different_columns:
            raise ValueError(f"These numerical columns are not found in the train data or "
                             f"their types were not numerical: {', '.join(different_columns)}.")
        for col, scaler in self._scaler_per_col.items():
            trans_data[col] = scaler.transform(trans_data[col].values.reshape(-1, 1))
        return trans_data

    @staticmethod
    def _get_scaler(strategy: ScalerType):
        """returns appropriate scaler according to scaling strategy"""
        if strategy is ScalerType.MINMAX:
            return MinMaxScaler()
        return StandardScaler()



"""Includes UnexpectedNominalValuesControl class."""
from typing import List, Set

from organon.fl.core.helpers.string_helper import is_null_or_empty
from organon.fl.logging.helpers.log_helper import LogHelper
from organon.idq.core.enums.data_entity_type import DataEntityType
from organon.idq.core.enums.dq_comparison_result_code import DqComparisonResultCode
from organon.idq.domain.businessobjects.dq_calculation_result import DqCalculationResult
from organon.idq.domain.businessobjects.dq_comparison_result import DqComparisonResult
from organon.idq.domain.businessobjects.statistics.dq_categorical_statistics import DqCategoricalStatistics
from organon.idq.domain.controls.base_dq_control import BaseDqControl
from organon.idq.domain.enums.dq_control_type import DqControlType
from organon.idq.domain.enums.dq_test_group_type import DqTestGroupType


class UnexpectedNominalValuesControl(BaseDqControl):
    """Control class for checking unexpected nominal values in data."""

    @classmethod
    def get_test_group_type(cls) -> DqTestGroupType:
        return DqTestGroupType.RULE_BASED_CONTROLS_COLUMN_SET

    @staticmethod
    def get_control_type() -> DqControlType:
        return DqControlType.UNEXPECTED_NOMINAL_VALUES

    def get_description(self) -> str:
        return "Unexpected Nominal Values"

    def _execute_control(self) -> List[DqComparisonResult]:
        results = []
        nominal_comp_cols = self.comp_context.get_nominal_comparison_columns()
        test_group_info = self.comp_context.test_group_info[self.get_test_group_type()]
        for col in nominal_comp_cols:
            col_name = col.column_name
            effective_bmh = min(test_group_info.test_bmh, col.benchmark_horizon)
            max_index = max(len(self.comp_context.control_results) - effective_bmh, 0)
            filtered_controls = list(reversed(self.comp_context.control_results[max_index:]))
            test_stat_collection = self.comp_context.test_calc_result.population_stats.nominal_statistics
            if test_stat_collection is None:
                LogHelper.info("Test statistics collection is null, column values tests skipped")
            elif len(filtered_controls) > 0 and col_name in test_stat_collection:
                union_of_control_values = self.__get_control_values(filtered_controls, col_name)
                test_nominal_stats = test_stat_collection[col_name]
                test_values = set(UnexpectedNominalValuesControl.__get_unique_values(test_nominal_stats))
                unexpected_values = list(test_values - union_of_control_values)
                results.extend(self.__get_unexpected_values_results(unexpected_values, col_name))
        return results

    def __get_unexpected_values_results(self, unexpected_values: List[str], column: str):
        msg = f"Following unexpected values are found the first time in data for the column {column}: "
        msg += "\n".join([val if not is_null_or_empty(val) else "NULL" for val in unexpected_values])
        results = []
        for val in unexpected_values:
            result = DqComparisonResult(
                data_entity=DataEntityType.COLUMN,
                data_entity_name=column,
                test_group=self.get_test_group_type(),
                result_code=DqComparisonResultCode.UNEXPECTED_NOMINAL_VALUES,
                property_code="UnexpectedValues",
                message=msg
            )
            result.property_key_nominal = val
            results.append(result)
        return results

    @staticmethod
    def __get_control_values(filtered_controls: List[DqCalculationResult], col_name: str) -> Set[str]:
        union_of_control_values: Set[str] = set()
        for control in filtered_controls:
            control_stats_collection = control.population_stats.nominal_statistics
            if col_name not in control_stats_collection:
                continue
            union_of_control_values.update(UnexpectedNominalValuesControl.__get_unique_values(
                control_stats_collection[col_name]))
        return union_of_control_values

    @staticmethod
    def __get_unique_values(categorical_stats: DqCategoricalStatistics):
        return [str(key) if key is not None else None for key in categorical_stats.frequencies.keys()]



"""
This module includes BaseAfeModelOutput class.
"""
from datetime import datetime
from typing import Dict, List, TypeVar, Generic

from organon.afe.domain.enums.afe_learning_type import AfeLearningType
from organon.afe.domain.modelling.businessobjects.base_afe_feature import BaseAfeFeature
from organon.afe.domain.modelling.businessobjects.transaction_file_stats import TransactionFileStats
from organon.afe.domain.reporting.runtime_statistics import RuntimeStatistics
from organon.afe.domain.reporting.transformation import Transformation
from organon.afe.domain.settings.afe_date_column import AfeDateColumn
from organon.afe.domain.settings.target_descriptor import TargetDescriptor

AfeFeatureType = TypeVar("AfeFeatureType", bound=BaseAfeFeature)


class BaseAfeModelOutput(Generic[AfeFeatureType]):
    """Base AFE output information of model"""

    def __init__(self):
        self.model_identifier: str = None
        self.build_date: datetime = None
        self.output_features: Dict[str, AfeFeatureType] = None
        self.transformations: Dict[str, Transformation] = None
        self.transaction_file_stats: TransactionFileStats = None
        self.runtime_stats: RuntimeStatistics = None
        self.trx_entity_column: str = None
        self.trx_date_columns: Dict[str, AfeDateColumn] = None
        self.target_descriptor: TargetDescriptor = None
        self.final_column_metrics: List[dict] = None
        self.afe_learning_type: AfeLearningType = None
        self.all_features: Dict[str, AfeFeatureType] = None



"""Includes GamMixin class."""
from typing import List, Dict, Any

import numpy as np
import pandas as pd
from sklearn.linear_model import Lasso
from sklearn.model_selection import GridSearchCV
from statsmodels.stats.outliers_influence import variance_inflation_factor

from organon.fl.core.exceptionhandling.known_exception import KnownException
from organon.ml.common.enums.target_type import TargetType
from organon.ml.preprocessing.services.preprocessor import Preprocessor


class GamMixin:
    """Mixin class including common gam functions"""

    def _fit_gam(self, train_data: pd.DataFrame, target_data: pd.Series):
        self._coarse_class = Preprocessor.get_coarse_class_service(0.2, target_column_type=self._get_target_type().name)
        self._coarse_class.fit(train_data, target_data)
        train_data = self._get_preprocessed(train_data)
        target_data = target_data.to_numpy(copy=False)
        self._initial_selected_features, self._best_alpha = self._get_initial_keep_list_and_best_alpha(train_data,
                                                                                                       target_data)
        keep_list = self._reduce_features(train_data, target_data)
        if len(keep_list) == 0:
            keep_list = train_data.columns.to_list()
        self._selected_features = keep_list
        self._model = self._fit_and_get_final_model(train_data, target_data, keep_list)
        return self._model

    def _get_preprocessed(self, data: pd.DataFrame):
        rejected_cols = self._coarse_class.fit_output.rejected_list
        if len(rejected_cols) > 0:
            data = data.drop(columns=[col for col in data.columns if col in rejected_cols])
        return self._coarse_class.transform(data)

    def _get_target_type(self) -> TargetType:
        raise NotImplementedError

    @property
    def final_model(self):
        """Returns final model"""
        return self._model

    @property
    def best_alpha(self):
        """Returns alpha value selected after GridSearchCV with Lasso"""
        return self._best_alpha

    @property
    def selected_features(self):
        """Returns column names selected after fit"""
        return self._selected_features

    @property
    def initial_selected_features(self):
        """Returns column names selected after GridSearchCV with Lasso"""
        return self._initial_selected_features

    def _fit_and_get_final_model(self, train_data: pd.DataFrame, target_data: np.ndarray, keep_list: List[str]):
        raise NotImplementedError

    def _get_initial_keep_list_and_best_alpha(self, train_data: pd.DataFrame, target_data: np.ndarray):
        param_grid = [{'alpha': self.grid_search_param_set, 'positive': [True]}]
        model = Lasso()
        grid_search = GridSearchCV(model, param_grid, cv=self.grid_search_cv, scoring=self.grid_search_scoring)
        grid_search.fit(train_data, target_data)

        lasso_coef_df_tmp = pd.DataFrame(grid_search.best_estimator_.coef_, columns=['coef'],
                                         index=train_data.columns).reset_index()
        keep_list = list(lasso_coef_df_tmp[lasso_coef_df_tmp['coef'] != 0]['index'].values)
        return keep_list, grid_search.best_params_['alpha']

    def _reduce_features(self, train_data: pd.DataFrame, target_data: np.ndarray) -> List[str]:
        raise NotImplementedError

    @classmethod
    def _get_feature_importances(cls, train_data: pd.DataFrame, target_data: np.ndarray):
        raise NotImplementedError

    @classmethod
    def _get_feature_to_remove(cls, train_data: pd.DataFrame, target_data: np.ndarray, keep_list: List[str],
                               candidate_features: List[str]):
        if len(candidate_features) == 1:
            return candidate_features[0]
        imp_df = cls._get_feature_importances(cls._get_cols(train_data, keep_list), target_data)
        imp_df.reset_index(inplace=True)
        imp_df.columns = ['feature', 'importance']
        imp_df = imp_df[imp_df["feature"].isin(candidate_features)].sort_values(by='importance')
        largest_vif_feature = imp_df.head(1)['feature'].values[0]
        return largest_vif_feature

    @classmethod
    def _get_keep_list(cls, train_data: pd.DataFrame, target_data: np.ndarray, keep_list_all: List[str],
                       best_alpha: float):
        if best_alpha == 0:
            model = Lasso(alpha=best_alpha, positive=True, max_iter=10000)
        else:
            model = Lasso(alpha=best_alpha, positive=True)
        data = cls._get_cols(train_data, keep_list_all)
        columns = data.columns.tolist()
        model.fit(data, target_data)
        del data
        lasso_coef_df_tmp = pd.DataFrame(model.coef_, columns=['coef'], index=columns).reset_index()
        keep_list = list(lasso_coef_df_tmp[lasso_coef_df_tmp['coef'] != 0]['index'].values)
        if len(keep_list) == 0:
            raise KnownException("No significant features found. Cannot build model.")
        return keep_list

    @classmethod
    def _get_vif_data(cls, train_data: pd.DataFrame, keep_list: List[str]):
        vif_data = pd.DataFrame()
        vif_data["feature"] = keep_list
        if len(keep_list) > 1:
            keep_df_arr = cls._get_cols(train_data, keep_list)
            vif_data["VIF"] = [variance_inflation_factor(keep_df_arr, i) for i in range(len(keep_list))]
        else:
            vif_data["VIF"] = [0]
        return vif_data

    @classmethod
    def _get_cols(cls, frame: pd.DataFrame, cols: List[str]):
        if sorted(frame.columns.tolist()) == sorted(cols):
            return frame  # prevents new copy of frame if all columns are wanted
        return frame[cols]

    def _get_params_with_defaults(self) -> Dict[str, Any]:
        # pylint: disable=no-self-use
        return {
            "grid_search_param_set": np.arange(0, 2, 0.05).tolist(),
            "grid_search_cv": 10,
            "grid_search_scoring": "neg_mean_squared_error",
            "max_vif_limit": 5
        }



""" This module includes UnivariatePerformanceFeatureReduction"""
from typing import List

from organon.ml.feature_reduction.domain.enums.feature_reduction_types import FeatureReductionType
from organon.ml.feature_reduction.domain.helpers.univariate_performance_helper import \
    get_univariate_performances_for_columns
from organon.ml.feature_reduction.domain.objects.feature_reduction_output import FeatureReductionOutput
from organon.ml.feature_reduction.domain.objects.univariate_performance_reduction_output import \
    UnivariatePerformanceFeatureReductionOutput
from organon.ml.feature_reduction.domain.reductions.base_feature_reduction import BaseFeatureReduction
from organon.ml.feature_reduction.settings.objects.univariate_performance_feature_reduction_settings import \
    UnivariatePerformanceFeatureReductionSettings


class UnivariatePerformanceFeatureReduction(BaseFeatureReduction):
    """UnivariatePerformanceFeatureReduction class"""

    def _execute_reduction(self, settings: UnivariatePerformanceFeatureReductionSettings) -> FeatureReductionOutput:
        included_columns = self._get_included_columns(settings.data, settings.excluded_columns)
        if settings.target_column_name in included_columns:
            included_columns.remove(settings.target_column_name)

        scores = get_univariate_performances_for_columns(
            settings.data, included_columns,
            settings.target_column_name, settings.target_type, settings.performance_metric,
            settings.random_state
        )
        if len(scores) == 0:
            reduced_columns = None
        else:
            reduced_columns = [col for col, score in scores.items()
                               if score is not None and score < settings.univariate_performance_threshold]
        output = UnivariatePerformanceFeatureReduction._get_output(reduced_columns, scores)
        return output

    @staticmethod
    def _get_output(reduced_columns: List[str], score_result) -> UnivariatePerformanceFeatureReductionOutput:
        """
        Get output for univariate performance feature reduction.
        Parameters
        ----------
        reduced_columns
        score_result

        Returns
        -------
        Returns UnivariatePerformanceFeatureReductionOutput
        """
        output = UnivariatePerformanceFeatureReductionOutput()
        output.feature_reduction_type = FeatureReductionType.UNIVARIATE_PERFORMANCE
        output.reduced_column_list = reduced_columns
        output.univariate_performance_result = score_result
        return output

    def get_description(self) -> str:
        return "Univariate Performance Feature Reduction"



"""Includes FeatureReductionUserInputService class."""
from typing import List

import pandas as pd
from organon.ml.common.helpers.user_input_service_helper import get_default_if_none, get_enum
from organon.fl.core.exceptionhandling.known_exception import KnownException
from organon.ml.common.enums.target_type import TargetType
from organon.ml.feature_reduction.domain.enums.feature_reduction_types import FeatureReductionType
from organon.ml.feature_reduction.settings.objects.base_feature_reduction_settings import BaseFeatureReductionSettings
from organon.ml.feature_reduction.settings.objects.high_correlated_feature_reduction_settings import \
    HighCorrelatedFeatureReductionSettings
from organon.ml.feature_reduction.settings.objects.null_feature_reduction_settings import NullFeatureReductionSettings
from organon.ml.feature_reduction.settings.objects.similar_distribution_feature_reduction_settings import \
    SimilarDistributionFeatureReductionSettings
from organon.ml.feature_reduction.settings.objects.stability_feature_reduction_settings import \
    StabilityFeatureReductionSettings
from organon.ml.feature_reduction.settings.objects.univariate_performance_feature_reduction_settings import \
    UnivariatePerformanceFeatureReductionSettings
from organon.ml.feature_reduction.settings.objects.user_feature_reduction_settings import UserFeatureReductionSettings

class FeatureReductionUserInputService:
    """Service for validating user input and generating settings for feature reduction"""

    @staticmethod
    def _set_default_metric(value: str, target_type: TargetType) -> str:
        """
        Set default metric
        :param value:
        :param target_type:
        :return:
        """
        if value is not None:
            return value
        if target_type.name == TargetType.BINARY.name:
            return "roc_auc"
        if target_type.name == TargetType.MULTICLASS.name:
            return "roc_auc_ovr_weighted"
        if target_type.name == TargetType.SCALAR.name:
            return "r2"
        raise KnownException("Please control target type.")

    @staticmethod
    def _set_default_univariate_performance_threshold(value: float, target_type: TargetType) -> float:
        """Set default univariate performance threshold"""
        if value is not None:
            return value
        if target_type.name == TargetType.BINARY.name:
            return 0.55
        if target_type.name == TargetType.MULTICLASS.name:
            return 0.55
        if target_type.name == TargetType.SCALAR.name:
            return 0.05
        raise KnownException("Please control target type.")

    @staticmethod
    def _get_included_reduction_types(input_settings: UserFeatureReductionSettings) -> List[str]:
        """
        Return included_reduction_types
        :param input_settings:
        :return:
        """
        included_reduction_types = input_settings.included_reduction_types
        if included_reduction_types is None:
            included_reduction_types_list = [FeatureReductionType.NULL.name,
                                             FeatureReductionType.STABILITY.name,
                                             FeatureReductionType.UNIVARIATE_PERFORMANCE.name,
                                             FeatureReductionType.SIMILAR_DISTRIBUTION.name,
                                             FeatureReductionType.HIGH_CORRELATION.name
                                             ]
            if input_settings.target_column_name is None:
                included_reduction_types_list.remove(FeatureReductionType.UNIVARIATE_PERFORMANCE.name)
        else:
            included_reduction_types_list = included_reduction_types
        return included_reduction_types_list

    def generate_feature_reduction_settings(self, input_settings: UserFeatureReductionSettings) -> List[
        BaseFeatureReductionSettings]:

        """
        Generate Feature Reduction Settings list
        :return:
        """
        settings = []
        if input_settings.target_column_name is not None:
            excluded_column_names = input_settings.excluded_columns + [
                input_settings.target_column_name] if input_settings.excluded_columns is not None else [
                input_settings.target_column_name]
        else:
            excluded_column_names = input_settings.excluded_columns
        included_reduction_types = self._get_included_reduction_types(input_settings)

        if FeatureReductionType.NULL.name in included_reduction_types:
            settings.append(self.get_null_feature_reduction_settings(
                input_settings.data, input_settings.null_ratio_threshold, excluded_column_names))

        if FeatureReductionType.STABILITY.name in included_reduction_types:
            settings.append(
                self.get_stability_feature_reduction_settings(input_settings.data, excluded_column_names))

        if FeatureReductionType.UNIVARIATE_PERFORMANCE.name in included_reduction_types:
            settings.append(self.get_univariate_performance_feature_reduction_settings(
                input_settings.data, input_settings.target_type, input_settings.target_column_name,
                input_settings.performance_metric, input_settings.univariate_performance_threshold,
                input_settings.random_state, excluded_column_names))

        if FeatureReductionType.SIMILAR_DISTRIBUTION.name in included_reduction_types:
            settings.append(self.get_similar_distribution_feature_reduction_settings(
                input_settings.data, input_settings.target_type, input_settings.target_column_name,
                input_settings.performance_metric, input_settings.nunique_count,
                input_settings.random_state, excluded_column_names))

        if FeatureReductionType.HIGH_CORRELATION.name in included_reduction_types:
            settings.append(self.get_high_correlated_feature_reduction_settings(
                input_settings.data, input_settings.target_type, input_settings.target_column_name,
                input_settings.performance_metric, input_settings.correlation_threshold,
                input_settings.random_state, excluded_column_names))

        return settings

    @classmethod
    def get_stability_feature_reduction_settings(cls, data: pd.DataFrame, excluded_col_names: List[
        str] = None) -> StabilityFeatureReductionSettings:
        """
        Validates settings for stability feature reduction entered by user and generates settings object
        :param data:
        :param excluded_col_names:
        :return:
        """
        if data is None:
            raise ValueError("Data should be given")
        return StabilityFeatureReductionSettings(data, excluded_columns=excluded_col_names)

    @classmethod
    def get_null_feature_reduction_settings(cls, data: pd.DataFrame, null_ratio_threshold: float = None,
                                            excluded_col_names: List[str] = None) -> NullFeatureReductionSettings:
        """
        Validates settings for null feature reduction entered by user and generates settings object
        :param data:
        :param null_ratio_threshold:
        :param excluded_col_names:
        :return:
        """
        if data is None:
            raise ValueError("Data should be given")
        null_ratio_threshold = get_default_if_none(null_ratio_threshold, 0.99)
        return NullFeatureReductionSettings(data, null_ratio_threshold, excluded_columns=excluded_col_names)

    @classmethod
    def get_high_correlated_feature_reduction_settings(cls, data: pd.DataFrame, target_type: str = None,
                                                       target_column_name: str = None, performance_metric: str = None,
                                                       correlation_threshold: float = None,
                                                       random_state=None, excluded_col_names: List[
                str] = None) -> \
            HighCorrelatedFeatureReductionSettings:
        """
        Validates settings for high correlated feature reduction entered by user and generates settings object
        :param data:
        :param target_type:
        :param target_column_name:
        :param performance_metric:
        :param correlation_threshold:
        :param random_state:
        :return:
        """
        # pylint: disable=too-many-arguments
        if data is None:
            raise ValueError("Data should be given")
        correlation_threshold = get_default_if_none(correlation_threshold, 0.99)
        target_type = get_enum(target_type, TargetType)
        if target_column_name is not None:
            if target_type is None:
                raise ValueError("Type of target column should be given.")
            performance_metric = FeatureReductionUserInputService._set_default_metric(performance_metric, target_type)
            random_state = get_default_if_none(random_state, 42)
        return HighCorrelatedFeatureReductionSettings(data, target_type, target_column_name,
                                                      performance_metric,
                                                      correlation_threshold, None, random_state,
                                                      excluded_columns=excluded_col_names)

    @classmethod
    def get_similar_distribution_feature_reduction_settings(cls, data: pd.DataFrame, target_type: str = None,
                                                            target_column_name: str = None,
                                                            performance_metric: str = None,
                                                            nunique_count: int = None,
                                                            random_state=None,
                                                            excluded_col_names: List[str] = None) -> \
            SimilarDistributionFeatureReductionSettings:
        """
        Validates settings for similar distribution feature reduction
        entered by user and generates settings object
        :param data:
        :param target_type:
        :param target_column_name:
        :param performance_metric:
        :param nunique_count:
        :param random_state:
        :return:
        """
        # pylint: disable=too-many-arguments
        if data is None:
            raise ValueError("Data should be given")
        target_type = get_enum(target_type, TargetType)
        nunique_count = get_default_if_none(nunique_count, 20)
        if target_column_name is not None:
            if target_type is None:
                raise ValueError("Type of target column should be given.")
            performance_metric = FeatureReductionUserInputService._set_default_metric(performance_metric, target_type)
            random_state = get_default_if_none(random_state, 42)

        return SimilarDistributionFeatureReductionSettings(data, target_type, target_column_name,
                                                           performance_metric,
                                                           nunique_count, None, random_state,
                                                           excluded_columns=excluded_col_names)

    @classmethod
    def get_univariate_performance_feature_reduction_settings(cls, data: pd.DataFrame, target_type: str,
                                                              target_column_name: str, performance_metric: str = None,
                                                              univariate_performance_threshold: float = None,
                                                              random_state=None, excluded_col_names: List[
                str] = None) -> UnivariatePerformanceFeatureReductionSettings:
        """
        Validates settings for univariate performance feature reduction entered by user and generates settings object
        :param data:
        :param target_type:
        :param target_column_name:
        :param performance_metric:
        :param univariate_performance_threshold:
        :param random_state:
        :return:
        """
        # pylint: disable=too-many-arguments
        if data is None:
            raise ValueError("Data should be given")
        if target_column_name is None or target_type is None:
            raise ValueError("Target column name and type of target column should be given.")
        target_type = get_enum(target_type, TargetType)
        performance_metric = FeatureReductionUserInputService._set_default_metric(performance_metric, target_type)
        random_state = get_default_if_none(random_state, 42)

        univariate_performance_threshold = FeatureReductionUserInputService \
            ._set_default_univariate_performance_threshold(univariate_performance_threshold, target_type)
        return UnivariatePerformanceFeatureReductionSettings(data, target_type, target_column_name,
                                                             performance_metric,
                                                             univariate_performance_threshold, random_state,
                                                             excluded_columns=excluded_col_names)



"""Includes Selecter class"""
from typing import List

import pandas as pd

from organon.ml.modelling.algorithms.core.abstractions.base_modeller import BaseModeller
from organon.ml.modelling.model_selection.domain.objects.selecter_output import SelecterOutput
from organon.ml.modelling.model_selection.domain.services.selection_service import SelectionService
from organon.ml.modelling.model_selection.services.user_settings.user_selection_settings import UserSelectionSettings
from organon.ml.modelling.model_selection.settings.model_selection_user_input_service import \
    ModelSelectionUserInputService


class Selecter:
    """Best model selecter"""

    def __init__(self, modellers: List[BaseModeller], train_data: pd.DataFrame, target_data: pd.Series, *,
                 target_type: str = None, cv_count: int = 5,
                 test_data: pd.DataFrame = None, test_target_data: pd.Series = None,
                 add_stacking=True, add_voting=True, num_threads_for_parallel_fit:int=None,
                 num_threads_for_parallel_cv:int=None):
        # pylint: disable=too-many-arguments
        user_settings = UserSelectionSettings(modellers, train_data, target_data, target_type,
                                              cv_count, test_data, test_target_data, add_stacking, add_voting,
                                              num_threads_for_parallel_fit, num_threads_for_parallel_cv)
        self.settings = ModelSelectionUserInputService.get_selection_settings(user_settings)
        self.output: SelecterOutput = None

    def select(self):
        """Executes selecter and returns output"""
        service = SelectionService()
        self.output = service.select(self.settings)
        return self.output



"""todo"""
from typing import Optional


class DqBaseStatistics:
    """todo"""
    def __init__(self):
        self.n_val: Optional[float] = None
        self.n_miss: Optional[float] = None
        self.cardinality: Optional[float] = None



"""
This module keeps the helper functions for double type.
"""
import decimal
import math
from decimal import Decimal

from organon.fl.mathematics.constants import DOUBLE_MAX, DOUBLE_MIN


def round_value(value: float, places: int) -> float:
    """
    Rounds the value.
    :param value: float value to be rounded
    :param places: decimal points
    :return: rounded float value
    """
    if places < 0:
        raise ValueError
    decimal_value = Decimal(value)
    decimal_value = decimal_value.quantize(Decimal('0.' + places * '0'), rounding=decimal.ROUND_HALF_UP)
    return float(decimal_value)


def is_extreme(value: float) -> bool:
    """Returns true if value has an extreme value like nan ,infinity etc."""
    if math.isnan(value) or value == float("inf") or value == float("-inf") or value <= DOUBLE_MIN \
            or value >= DOUBLE_MAX:
        return True
    return False


def equals(val1: float, val2: float):
    """Returns true if both vals equal or both nan"""
    if math.isnan(val1):
        return math.isnan(val2)
    return val1 == val2



"""
This module includes enum class AfeOperator.
"""
from enum import Enum
from typing import List


# pylint: disable=invalid-name
class AfeOperator(Enum):
    """
    Operators for AFE calculations
    """
    Density = 0
    Min = 1
    Max = 2
    Sum = 3
    Frequency = 4
    TimeSinceFirst = 5
    TimeSinceLast = 6
    Ratio = 7
    SumTrend = 8
    FrequencyTrend = 9
    DensityTrend = 10
    CountDistinct = 11
    Mode = 12


OPERATOR_SET_ONE: List[AfeOperator] = [AfeOperator.Density, AfeOperator.Sum, AfeOperator.Max,
                                       AfeOperator.Min, AfeOperator.Frequency, AfeOperator.CountDistinct,
                                       AfeOperator.Mode]
OPERATOR_SET_TWO: List[AfeOperator] = [AfeOperator.TimeSinceFirst, AfeOperator.TimeSinceLast]
OPERATOR_SET_THREE: List[AfeOperator] = [AfeOperator.SumTrend, AfeOperator.DensityTrend,
                                         AfeOperator.FrequencyTrend]



"""Includes helpers for object classification"""
from organon.fl.core.fileoperations import directory_helper
from organon.fl.logging.helpers.log_helper import LogHelper


def check_directory(directory: str):
    """Checks if given directory exists"""
    if not directory_helper.exists(directory):
        throw_val_ex_with_log(f"Directory: {directory} not found!")


def throw_val_ex_with_log(msg: str):
    """Raises and logs a value error """
    LogHelper.error(msg)
    raise ValueError(msg)



"""This module includes "Interval" class."""
import functools

from organon.fl.core.exceptionhandling.known_exception import KnownException
from organon.fl.mathematics.enums.sets.interval_type import IntervalType


@functools.total_ordering
class Interval:
    """Class for Interval"""

    def __init__(self, interval_id: str, lower_bound: float, upper_bound: float, interval_type: IntervalType):
        self.__interval_id: str = interval_id
        self.__lower_bound: float = lower_bound
        self.__upper_bound: float = upper_bound
        self.__interval_type: IntervalType = interval_type
        self._func = self.__get_containment_function(self.__interval_type)

    @property
    def interval_id(self) -> str:
        """Returns interval_id of Interval."""
        return self.__interval_id

    @property
    def lower_bound(self) -> float:
        """Returns lower_bound of Interval."""
        return self.__lower_bound

    @property
    def upper_bound(self) -> float:
        """Returns upper_bound of Interval."""
        return self.__upper_bound

    @property
    def interval_type(self) -> IntervalType:
        """Returns interval_type of Interval."""
        return self.__interval_type

    @property
    def func(self):
        """Returns func of Interval."""
        return self._func

    @property
    def is_degenerate(self) -> bool:
        """Control for degenerate"""
        return (self.lower_bound <= self.upper_bound) & (self.lower_bound >= self.upper_bound)

    def __get_containment_function(self, interval_type: IntervalType):

        if interval_type == IntervalType.CLOSED_CLOSED:
            return lambda s: ((s >= self.lower_bound) & (s <= self.upper_bound))
        if interval_type == IntervalType.OPEN_CLOSED:
            return lambda s: ((s > self.lower_bound) & (s <= self.upper_bound))
        if interval_type == IntervalType.CLOSED_OPEN:
            return lambda s: ((s >= self.lower_bound) & (s < self.upper_bound))
        if interval_type == IntervalType.OPEN_OPEN:
            return lambda s: ((s > self.lower_bound) & (s < self.upper_bound))

        raise KnownException("Wrong Interval Type")

    def contains(self, value: float):
        """Check for contains"""
        return self.func(value)

    def deep_copy(self):
        """Deep copy for Interval object"""
        return Interval(self.interval_id, self.lower_bound, self.upper_bound, self.interval_type)

    def to_string(self) -> str:
        """to_string method for Interval object"""
        if self.interval_type == IntervalType.CLOSED_CLOSED:
            return f'[{self.lower_bound},{self.upper_bound}]'
        if self.interval_type == IntervalType.CLOSED_OPEN:
            return f'[{self.lower_bound},{self.upper_bound})'
        if self.interval_type == IntervalType.OPEN_CLOSED:
            return f'({self.lower_bound},{self.upper_bound}]'
        if self.interval_type == IntervalType.OPEN_OPEN:
            return f'({self.lower_bound},{self.upper_bound})'
        return ""

    def __eq__(self, interval: "Interval"):
        """Compare interval"""
        if self.is_degenerate == interval.is_degenerate:
            return self.lower_bound == interval.lower_bound
        return False

    def __hash__(self):
        return hash(self.lower_bound)

    def __lt__(self, interval: "Interval"):
        """Compare Interval"""
        if self.is_degenerate == interval.is_degenerate:
            return self.lower_bound < interval.lower_bound
        if interval.is_degenerate:
            return True
        return False

    def __getstate__(self):
        state = self.__dict__.copy()
        # Remove the unpicklable entries.
        del state['_func']
        return state

    def __setstate__(self, state):
        self.__dict__.update(state)
        self._func = self.__get_containment_function(self.__interval_type)



"""Includes UnexpectedNumericalValuesControl class."""
import math
from typing import List

from organon.fl.core.enums.column_native_type import ColumnNativeType
from organon.idq.core.enums.data_entity_type import DataEntityType
from organon.idq.core.enums.dq_comparison_result_code import DqComparisonResultCode
from organon.idq.core.helpers.formatting_helper import FormattingHelper
from organon.idq.domain.businessobjects.dq_calculation_result import DqCalculationResult
from organon.idq.domain.businessobjects.dq_comparison_result import DqComparisonResult
from organon.idq.domain.businessobjects.statistics.dq_sample_numerical_statistics import DqSampleNumericalStatistics
from organon.idq.domain.controls.base_dq_control import BaseDqControl
from organon.idq.domain.enums.dq_control_type import DqControlType
from organon.idq.domain.enums.dq_test_group_type import DqTestGroupType


class UnexpectedNumericalValuesControl(BaseDqControl):
    """Control class for checking unexpected min or max values in sample data."""

    @classmethod
    def get_test_group_type(cls) -> DqTestGroupType:
        return DqTestGroupType.RULE_BASED_CONTROLS_COLUMN_SET

    @staticmethod
    def get_control_type() -> DqControlType:
        return DqControlType.UNEXPECTED_NUMERICAL_VALUES

    def get_description(self) -> str:
        return "Unexpected Numerical Values"

    def _execute_control(self) -> List[DqComparisonResult]:
        results = []
        numerical_comp_cols = self.comp_context.get_comparison_cols_by_native_type(ColumnNativeType.Numeric)
        test_group_info = self.comp_context.test_group_info[self.get_test_group_type()]
        for col in numerical_comp_cols:
            col_name = col.column_name
            effective_bmh = min(test_group_info.test_bmh, col.benchmark_horizon)
            max_index = max(len(self.comp_context.control_results) - effective_bmh, 0)
            filtered_controls = list(reversed(self.comp_context.control_results[max_index:]))
            test_stat_collection = self.comp_context.test_calc_result.sample_stats.numerical_statistics
            if len(filtered_controls) > 0 and col_name in test_stat_collection:
                control_min, control_max = self.__get_control_min_max_values(filtered_controls, col_name)
                test_numerical_stats = test_stat_collection[col_name]
                self.__check_unexpected_min(results, test_numerical_stats, control_min, col_name)
                self.__check_unexpected_max(results, test_numerical_stats, control_max, col_name)
        return results

    def __check_unexpected_min(self, results: List[DqComparisonResult],
                               test_numerical_stats: DqSampleNumericalStatistics, control_min_value: float,
                               col_name: str):
        test_min = test_numerical_stats.min
        is_less_than_min = test_min < control_min_value if not math.isnan(control_min_value) else True
        if not math.isnan(test_min) and is_less_than_min:
            result = self._get_comparison_result(
                data_entity=DataEntityType.COLUMN,
                data_entity_name=col_name,
                result_code=DqComparisonResultCode.UNEXPECTED_MINIMUM,
                property_code="UnexpectedMinimum",
                message=f"The minimum value in the test set ({test_min}) is smaller than the control "
                        f"minimum ({control_min_value}) for the column {col_name}")
            result.property_key_numeric = FormattingHelper.to_nullable(test_min)
            results.append(result)

    def __check_unexpected_max(self, results: List[DqComparisonResult], test_numerical_stats, control_max_value: float,
                               col_name: str):
        test_max = test_numerical_stats.max
        is_higher_than_max = test_max > control_max_value if not math.isnan(control_max_value) else True
        if not math.isnan(test_max) and is_higher_than_max:
            result = self._get_comparison_result(
                data_entity=DataEntityType.COLUMN,
                data_entity_name=col_name,
                result_code=DqComparisonResultCode.UNEXPECTED_MAXIMUM,
                property_code="TestMaximum",
                message=f"The maximum value in the test set ({test_max}) is greater than the control "
                        f"maximum ({control_max_value}) for the column {col_name}")
            result.property_key_numeric = FormattingHelper.to_nullable(test_max)
            results.append(result)

    @staticmethod
    def __get_control_min_max_values(filtered_controls: List[DqCalculationResult], col_name: str):
        min_values, max_values = [], []
        for control in filtered_controls:
            stat_collection = control.sample_stats.numerical_statistics
            if col_name not in stat_collection:
                continue
            numerical_stats = stat_collection[col_name]
            current_min, current_max = numerical_stats.min, numerical_stats.max
            if not math.isnan(current_min):
                min_values.append(current_min)
            if not math.isnan(current_max):
                max_values.append(current_max)
        if len(min_values) == 0 and len(max_values) == 0:
            return float("nan"), float("nan")

        return min(min_values), max(max_values)



"""This module includes DataFrameOperationsFactory class."""
from organon.fl.core.businessobjects.dask_dataframe import DaskDataFrame
from organon.fl.core.businessobjects.dataframe import DataFrame
from organon.fl.core.businessobjects.dict_dataframe import DictDataFrame
from organon.fl.core.businessobjects.idataframe import IDataFrame
from organon.fl.mathematics.helpers.dask_dataframe_operations import DaskDataFrameOperations
from organon.fl.mathematics.helpers.dict_dataframe_operations import DictDataFrameOperations
from organon.fl.mathematics.helpers.pandas_dataframe_operations import PandasDataFrameOperations


class DataFrameOperationsFactory:
    """Class for DataFrameOperationsFactory"""

    @staticmethod
    def get_dataframe_operations(df_obj: IDataFrame):
        """ check whether we have dask or not and return dask operations accordingly"""
        if isinstance(df_obj, DictDataFrame):
            return DictDataFrameOperations()
        if isinstance(df_obj, DataFrame):
            return PandasDataFrameOperations()
        if isinstance(df_obj, DaskDataFrame):
            return DaskDataFrameOperations()
        raise NotImplementedError



"""
This module includes enum class AfeDateColumnEncodingType.
"""
from enum import Enum


# pylint: disable=invalid-name
class AfeDateColumnEncodingType(Enum):
    """
    Keeps encoding type for a column.
    Datetime: datetime.datetime type is an example.
    String: a date column can be a string like '20221001'
    Integer: a date column can be a number like 20220101
    """
    DateTime = 0
    String = 1
    Integer = 2



"""Includes UserDbObjLocator class."""


class UserDbObjLocator:
    """Class for db object location definition in user input"""

    def __init__(self):
        self.connection_name: str = None
        self.table_name: str = None
        self.schema_name: str = None
        self.sql_statement: str = None



"""Includes DqDfComparisonParameters class."""
from organon.idq.domain.settings.abstractions.dq_base_comparison_parameters import DqBaseComparisonParameters
from organon.idq.domain.settings.input_source.dq_df_input_source_settings import DqDfInputSourceSettings


class DqDfComparisonParameters(DqBaseComparisonParameters[DqDfInputSourceSettings]):
    """DQ comparison parameters for dataframe source type"""



"""
This module includes TransactionFileRecord class.
"""
import numpy as np


class TransactionFileRecord:
    """
    Class corresponding to a row in a transaction file
    """

    def __init__(self, date: float, q_array: np.ndarray, d_array: np.ndarray):
        self.__date: float = date
        self.__q_array: np.ndarray = q_array
        self.__d_array: np.ndarray = d_array

    @property
    def date(self) -> float:
        """Returns value of private attribute '__date'"""
        return self.__date

    def get_quantity(self, index: int) -> np.float64:
        """Returns value of element in given index of private list '__q_array'"""
        return self.__q_array[index]

    def get_dimension(self, index: int) -> np.short:
        """Returns value of element in given index of private list '__d_array'"""
        return self.__d_array[index]

    def __lt__(self, other) -> bool:
        """Allows 'less than' comparison between TransactionFileRecord instances based on __date attribute"""
        return not self.date < other.date



"""Includes TrafficLightComparisonInput class."""
from organon.idq.core.enums.dq_comparison_result_code import DqComparisonResultCode
from organon.idq.domain.algorithms.objects.base_comparison_input import BaseComparisonInput


class TrafficLightComparisonInput(BaseComparisonInput):
    """Input dto for traffic light comparisons"""

    def __init__(self):
        super().__init__()
        self.green_threshold: float = None
        self.yellow_threshold: float = None
        self.result_code: DqComparisonResultCode = None
        self.property_code: str = None
        self.message: str = None



"""This module includes RuntimeStatsService class."""
import os
import time
from datetime import datetime
from threading import Thread

import psutil

from organon.afe.domain.reporting.runtime_statistics import RuntimeStatistics
from organon.fl.core.exceptionhandling.known_exception import KnownException
from organon.fl.core.executionutil.objects.stopwatch import StopWatch
from organon.fl.core.helpers import process_info_helper


class RuntimeStatsService:
    """Class for generating runtime statistics."""

    def __init__(self):
        self.__runtime_stats_thread: Thread = None
        self.__runtime_stats_thread_running = False
        self.__runtime_statistics: RuntimeStatistics = RuntimeStatistics()
        self.__watch: StopWatch = StopWatch(with_cpu_times=True)

    @property
    def runtime_statistics(self):
        """returns recorded runtime statistics"""
        return self.__runtime_statistics

    @property
    def is_running(self):
        """Returns if service is started or not"""
        return self.__runtime_stats_thread_running

    @property
    def is_thread_alive(self):
        """Returns if runtime stats thread is alive"""
        return self.__runtime_stats_thread.is_alive() if self.__runtime_stats_thread is not None else False

    def __record_runtime_stats(self, interval_in_seconds: int):
        process = psutil.Process(os.getpid())
        while self.__runtime_stats_thread_running:
            cpu_perc = process_info_helper.get_cpu_percentage_since_last_call(process)
            mem_usage = process_info_helper.get_memory_usage()
            stat_time = datetime.now()
            self.__runtime_statistics.cpu_usage[stat_time] = cpu_perc
            self.__runtime_statistics.memory_usage[stat_time] = mem_usage / (1024 * 1024)
            time.sleep(interval_in_seconds)

    def start_recording_runtime_stats(self, interval_in_seconds: float = 1):
        """Starts recording runtime statistics"""
        if self.__runtime_stats_thread_running:
            raise KnownException("Runtime stats service is already running")
        self.__runtime_stats_thread = Thread(target=self.__record_runtime_stats, args=(interval_in_seconds,))
        self.__runtime_stats_thread_running = True
        self.__runtime_stats_thread.start()
        self.__watch.start()

    def stop_recording_runtime_stats(self):
        """Stops recording runtime statistics"""
        if self.__runtime_stats_thread_running:
            self.__runtime_stats_thread_running = False
            self.__runtime_stats_thread.join()
            self.__runtime_statistics.execution_time = self.__watch.get_elapsed_seconds()
            cpu_times = self.__watch.get_cpu_times()
            self.__runtime_statistics.user_cpu_time = cpu_times[0]
            self.__runtime_statistics.system_cpu_time = cpu_times[1]

    def add_event(self, key, event_time=None):
        """Adds event to runtime stats"""
        if event_time is None:
            event_time = datetime.now()
        self.__runtime_statistics.event_times[key] = event_time



"""Includes UserInputSourceSettings class."""
from typing import Union, List, Callable

import pandas as pd

from organon.idq.services.user_settings.user_db_obj_locator import UserDbObjLocator


class UserInputSourceSettings:
    """Input source settings user class"""

    def __init__(self):
        self.source: Union[UserDbObjLocator, str, pd.DataFrame] = None
        self.name: str = None
        self.sampling_ratio: float = None
        self.max_num_of_samples: int = None
        self.number_of_rows_per_step: int = None
        self.csv_separator: str = None
        self.where_clause: str = None
        self.date_columns: List[str] = None
        self.filter_callable: Callable = None



"""Includes RecordSourceReader class."""
from typing import Tuple

import numpy as np
import pandas as pd

from organon.fl.core.exceptionhandling.known_exception import KnownException
from organon.idq.dataaccess.helpers.record_source_filter_helper import RecordSourceFilterHelper, FilteredTextFileReader
from organon.idq.domain.businessobjects.dq_data_source import DqDataSource
from organon.idq.domain.enums.dq_record_source_type import DqRecordSourceType
from organon.idq.domain.settings.abstractions.dq_base_input_source_settings import DqBaseInputSourceSettings
from organon.idq.domain.settings.input_source.dq_df_input_source_settings import DqDfInputSourceSettings
from organon.idq.domain.settings.input_source.dq_file_input_source_settings import DqFileInputSourceSettings


class RecordSourceReader:
    """
    Source reader class implementation for csv files, data frames, and db sources.
    """

    def __init__(self, data_source_settings: DqBaseInputSourceSettings, row_count: int = None):
        self.data_source_settings = data_source_settings
        self.row_count = row_count

    def read(self) -> DqDataSource:
        """
        Reads the table of records and returns a collection
        :return: A collection containing records in record source
        """
        source_type = self.data_source_settings.source.get_type()
        if source_type == DqRecordSourceType.TEXT:
            return self.read_text_source()
        if source_type == DqRecordSourceType.DATA_FRAME:
            return self.read_dataframe_source()
        raise KnownException("Record source type is not valid!")

    def read_text_source(self) -> DqDataSource:
        """
        Reads text source and returns a dq data source
        """
        if not isinstance(self.data_source_settings, DqFileInputSourceSettings):
            raise ValueError('Data source settings should be of type DqFileInputSourceSettings.')

        record_source: str = self.data_source_settings.source.locator
        included_columns = self.data_source_settings.included_columns  # included columns
        max_sample_size = self.data_source_settings.max_num_of_samples
        sep = self.data_source_settings.csv_separator
        columns = pd.read_csv(record_source, header=None, sep=sep, nrows=1).iloc[0].tolist()
        self._validate_source(included_columns, columns)

        dfs = RecordSourceFilterHelper.read_csv_with_chunks(self.data_source_settings)
        frame_chunks, full_row_count = self.read_df_chunks(dfs, max_sample_size)

        data_frame = pd.concat(frame_chunks)
        data_frame.reset_index(inplace=True, drop=True)
        data_frame = data_frame.convert_dtypes()
        dq_source = DqDataSource()
        dq_source.sampled_data = data_frame
        dq_source.full_data_row_count = full_row_count
        return dq_source

    def read_dataframe_source(self) -> DqDataSource:
        """Reads DataFrame source and returns a collection"""
        if not isinstance(self.data_source_settings, DqDfInputSourceSettings):
            raise ValueError('Data source settings should be of type DqDfInputSourceSettings.')

        ratio = self.data_source_settings.sampling_ratio
        max_sample_size = self.data_source_settings.max_num_of_samples
        source: pd.DataFrame = RecordSourceFilterHelper.get_filtered_dataframe(self.data_source_settings.source.locator,
                                                                               self.data_source_settings)
        full_row_count = len(source)
        included_columns: list = self.data_source_settings.included_columns

        self._validate_source(included_columns, list(source.columns))
        source_size = self.row_count if self.row_count is not None else len(source)
        if max_sample_size is not None and max_sample_size < source_size * ratio:
            if included_columns is None:
                data_frame: pd.DataFrame = source.sample(n=max_sample_size, replace=True,
                                                         random_state=1)
            else:
                data_frame: pd.DataFrame = source[included_columns].sample(n=max_sample_size, replace=True,
                                                                           random_state=1)
        elif ratio == 1.0:
            data_frame = source if included_columns is None else source[included_columns]
        else:
            if included_columns is None:
                data_frame: pd.DataFrame = source.sample(frac=ratio, replace=True, random_state=1)
            else:
                data_frame: pd.DataFrame = source[included_columns].sample(frac=ratio, replace=True, random_state=1)

        data_frame.reset_index(inplace=True, drop=True)

        dq_source = DqDataSource()
        dq_source.sampled_data = data_frame
        dq_source.full_data_row_count = full_row_count
        return dq_source

    def read_df_chunks(self, dfs: FilteredTextFileReader, max_sample_size: int) -> Tuple[list, int]:
        """

        :param self:
        :param dfs:
        :param max_sample_size:
        :return:
        """
        frame_chunks: list = []
        sample_count: int = 0
        is_full: bool = False
        full_row_count = 0
        for filtered_df in dfs:
            full_row_count += len(filtered_df)
            sample = filtered_df.sample(frac=self.data_source_settings.sampling_ratio)
            sample_size = len(sample)
            if sample_size + sample_count > max_sample_size:
                to_read = sample_size + sample_count - max_sample_size
                add_indices = np.random.choice(sample.index, to_read, replace=False)
                sample = sample[sample.index.isin(add_indices)]
                is_full = True
            frame_chunks.append(sample)
            sample_count += len(sample)
            if is_full:
                break

        return frame_chunks, full_row_count

    @staticmethod
    def _validate_source(included_columns: list, df_cols: list):
        """
        throws exception if source has duplicated columns
        or included_columns is not a subset of df_cols.
        :param included_columns: columns that we want to use
        :param df_cols: source columns
        :return:
        """
        RecordSourceReader.__check_included_cols(included_columns, df_cols)
        RecordSourceReader.__check_duplicated_cols(df_cols)

    @staticmethod
    def __check_included_cols(included_columns: list, df_cols: list):
        """
        throws exception if included_columns is not a subset of df_cols.
        :param included_columns: columns that we want to use
        :param df_cols: source columns
        :return:
        """
        if included_columns is not None:
            column_diff = set(included_columns) - set(df_cols)
            if len(column_diff) > 0:
                col_message_str = ""
                for col in column_diff:
                    col_message_str += col + ", "
                col_message_str = col_message_str[:-2]
                raise ValueError(f'These columns do not belong to the source: {col_message_str}.')

    @staticmethod
    def __check_duplicated_cols(df_cols: list):
        """
        throws exception if there are duplicated columns in df_cols.
        :param df_cols: source columns
        :return:
        """
        seen = set()
        dupes = [x for x in list(df_cols) if x in seen or seen.add(x)]
        if len(dupes) > 0:
            col_message_str = ""
            for dup_column in dupes:
                col_message_str += dup_column + ", "
            col_message_str = col_message_str[:-2]
            raise ValueError(f'These columns are duplicated in the source: {col_message_str}.')



"""Includes DqConstants class."""
from organon.fl.core.enums.column_native_type import ColumnNativeType
from organon.fl.core.exceptionhandling.known_exception import KnownException
from organon.fl.logging.helpers.log_helper import LogHelper
from organon.fl.security.helpers.encryption_helper import EncryptionHelper


class DqConstants:
    """Stores dq constants"""

    DEFAULT_PERCENTILE_LIST = [0.05, 1, 5, 10, 20, 25, 30, 40, 50, 60, 70, 75, 80, 90, 95, 99, 99.5]
    NUMERICAL_STATISTICS_MIN_CARDINALITY = 50
    NUMERICAL_STATISTICS_MAX_NUM_OF_INTERVALS = 20
    NUMERICAL_STATISTICS_MIN_INTERVAL_SIZE = 250
    OUTLIER_PARAM_DEFAULT = 3
    CALCULATION_LIMIT_DEFAULT = 5
    MAX_NOMINAL_CARDINALITY_COUNT_DEFAULT = 100
    CONTROL_MAXIMUM_NOM_CARDINALITY_DEFAULT = 1000
    TRAFFIC_LIGHT_THRESHOLD_GREEN = 0.2
    TRAFFIC_LIGHT_THRESHOLD_YELLOW = 0.4
    PSI_THRESHOLD_GREEN = 0.1
    PSI_THRESHOLD_YELLOW = 0.2
    Z_SCORE_DEFAULT = 1.96
    MINIMUM_CARDINALITY_DEFAULT = 1
    MINIMUM_INJECTION = 5
    NOMINAL_COLUMN_TYPES = [ColumnNativeType.String, ColumnNativeType.Date]
    __SECRET_KEY_DEFAULT = "qwertyuıop"  # nosec
    __secret_key: str = __SECRET_KEY_DEFAULT

    @staticmethod
    def get_secret_key():
        """Returns value of secret key"""
        secret_key = DqConstants.__secret_key
        if secret_key == DqConstants.__SECRET_KEY_DEFAULT:
            LogHelper.warning("Secret key not set. Will use default secret key. Set secret key for security.")
        return secret_key

    @staticmethod
    def set_secret_key(secret_key: str):
        """Sets global encryption secret key for AFE"""
        if len(secret_key) > EncryptionHelper.SECRET_KEY_MAX_LENGTH:
            raise KnownException(f"secret_key maximum length is {EncryptionHelper.SECRET_KEY_MAX_LENGTH}")
        DqConstants.__secret_key = secret_key



"""todo"""
from typing import List

from organon.fl.core.exceptionhandling.known_exception import KnownException
from organon.idq.core.enums.data_entity_type import DataEntityType
from organon.idq.core.enums.dq_comparison_result_code import DqComparisonResultCode
from organon.idq.domain.businessobjects.data_column.dq_data_column import DqDataColumn
from organon.idq.domain.businessobjects.data_column.dq_df_data_column import DqDfDataColumn
from organon.idq.domain.businessobjects.data_column.dq_file_data_column import DqFileDataColumn
from organon.idq.domain.businessobjects.dq_comparison_result import DqComparisonResult
from organon.idq.domain.businessobjects.dq_data_column_collection import DqDataColumnCollection
from organon.idq.domain.controls.base_dq_control import BaseDqControl
from organon.idq.domain.enums.dq_control_type import DqControlType
from organon.idq.domain.enums.dq_test_group_type import DqTestGroupType


class TableColumnsControl(BaseDqControl):
    """Control class for checking if table columns changed, added or removed"""

    @classmethod
    def get_test_group_type(cls) -> DqTestGroupType:
        return DqTestGroupType.TABLE_SCHEMA_CONTROLS

    @staticmethod
    def get_control_type() -> DqControlType:
        return DqControlType.TABLE_COLUMNS

    def get_description(self) -> str:
        return "Table Columns Control"

    def _execute_control(self) -> List[DqComparisonResult]:
        test_bmh = self.comp_context.test_group_info[self.get_test_group_type()].test_bmh
        if test_bmh < 1:
            raise ValueError("Bmh value cannot be less than 1")
        control_results = self.comp_context.control_results
        if len(control_results) == 0:
            raise ValueError("Benchmark result collection can not be empty")

        if self.comp_context.test_calc_result.data_source_stats is None:
            raise KnownException("T calculation result does not contain data source statistics.")

        max_index = max(len(control_results) - test_bmh, 0)
        filtered_controls = control_results[max_index:]
        calc_names = [calc_param.calculation_name for calc_param in self.comp_context.control_parameters]
        filtered_calc_names = calc_names[max_index:]
        test_col_collection = self.comp_context.test_calc_result.data_source_stats.data_column_collection
        results = []

        for i, control in enumerate(filtered_controls):
            if control.data_source_stats is None:
                raise KnownException(
                    f"Calculation {filtered_calc_names[i]} does not contain data source statistics, "
                    f"re-calculation may be required")
            control_col_collection = control.data_source_stats.data_column_collection
            results.extend(self.__compare_col_metadata(control_col_collection, test_col_collection,
                                                       filtered_calc_names[i]))
        return results

    @classmethod
    def __compare_col_metadata(cls, control_collection: DqDataColumnCollection, test_collection: DqDataColumnCollection,
                               calculation_name: str):
        results = []
        for control_col in control_collection:
            col_name = control_col.column_name
            test_col = test_collection.get_column(col_name)

            if test_col is None:
                comp_result = cls._get_comparison_result(
                    data_entity=DataEntityType.COLUMN,
                    data_entity_name=control_col.column_name,
                    result_code=DqComparisonResultCode.COLUMN_REMOVED,
                    property_code="RemovedColumnName",
                    message=f"Column {control_col.column_name} is removed"
                )
                comp_result.property_key_nominal = control_col.column_name
                comp_result.property_value_nominal = calculation_name
                results.append(comp_result)

            else:
                results.extend(cls._get_column_change_results(control_col, test_col, calculation_name))

        for col in test_collection:
            col_name = col.column_name
            control_col = control_collection.get_column(col_name)
            if control_col is not None:
                continue
            comp_result = cls._get_comparison_result(
                data_entity=DataEntityType.COLUMN,
                data_entity_name=col_name,
                result_code=DqComparisonResultCode.COLUMN_ADDED,
                property_code="AddedColumnName",
                message=f"Column {col_name} is added"
            )
            comp_result.property_key_nominal = col_name
            comp_result.property_value_nominal = calculation_name
            results.append(comp_result)
        return results

    @classmethod
    def _get_column_change_results(cls, control_col: DqDataColumn, test_col: DqDataColumn, calculation_name: str):
        if isinstance(control_col, DqDfDataColumn) and isinstance(test_col, DqDfDataColumn):
            return TableColumnsControl.__get_column_change_results_for_df_column(control_col, test_col,
                                                                                 calculation_name)
        if isinstance(control_col, DqFileDataColumn) and isinstance(test_col, DqFileDataColumn):
            return TableColumnsControl.__get_column_change_results_for_file_column(control_col, test_col,
                                                                                   calculation_name)
        raise NotImplementedError

    @staticmethod
    def __get_column_change_results_for_file_column(control_col: DqFileDataColumn, test_col: DqFileDataColumn,
                                                    calculation_name: str) -> List[DqComparisonResult]:
        message = ""

        if control_col.col_np_dtype != test_col.col_np_dtype:
            message += f"Control Column Data Type: {control_col.col_np_dtype}, " \
                       f"Test Column Data Type: {test_col.col_np_dtype}\n"

        return TableColumnsControl._get_col_type_changed_results(control_col, test_col, message, calculation_name)

    @staticmethod
    def __get_column_change_results_for_df_column(control_col: DqDfDataColumn, test_col: DqDfDataColumn,
                                                  calculation_name: str) -> List[DqComparisonResult]:
        message = ""

        if control_col.col_np_dtype != test_col.col_np_dtype:
            message += f"Control Column Data Type: {control_col.col_np_dtype}, " \
                       f"Test Column Data Type: {test_col.col_np_dtype}\n"

        return TableColumnsControl._get_col_type_changed_results(control_col, test_col, message, calculation_name)

    @classmethod
    def _get_col_type_changed_results(cls, control_col: DqDataColumn, test_col, message: str, calculation_name: str):
        results = []
        if message != "":
            comp_result1 = cls._get_comparison_result(
                data_entity=DataEntityType.COLUMN,
                data_entity_name=control_col.column_name,
                result_code=DqComparisonResultCode.COLUMN_TYPE_CHANGED,
                property_code="PreviousColumn",
                message=message
            )
            comp_result1.property_key_nominal = control_col.column_name
            comp_result1.property_value_nominal = calculation_name
            results.append(comp_result1)
            comp_result2 = cls._get_comparison_result(
                data_entity=DataEntityType.COLUMN,
                data_entity_name=control_col.column_name,
                result_code=DqComparisonResultCode.COLUMN_TYPE_CHANGED,
                property_code="CurrentColumn",
                message=message
            )
            comp_result2.property_key_nominal = test_col.column_name
            comp_result2.property_value_nominal = calculation_name
            results.append(comp_result2)
        return results



""" This module includes NullFeatureReduction"""
from typing import List

import pandas as pd

from organon.ml.feature_reduction.domain.enums.feature_reduction_types import FeatureReductionType
from organon.ml.feature_reduction.domain.objects.feature_reduction_output import FeatureReductionOutput
from organon.ml.feature_reduction.domain.reductions.base_feature_reduction import BaseFeatureReduction
from organon.ml.feature_reduction.settings.objects.null_feature_reduction_settings import NullFeatureReductionSettings


class NullFeatureReduction(BaseFeatureReduction):
    """ NullFeatureReduction class"""

    def _execute_reduction(self, settings: NullFeatureReductionSettings) -> FeatureReductionOutput:
        output = FeatureReductionOutput()
        included_columns = self._get_included_columns(settings.data, settings.excluded_columns)
        missing_df = self._calculate_null_ratio(settings.data, included_columns)
        output.feature_reduction_type = FeatureReductionType.NULL

        if missing_df is None:
            output.reduced_column_list = None
            return output

        reduced_columns = self._find_reduced_columns(missing_df, settings.null_ratio_threshold)
        output.reduced_column_list = reduced_columns
        return output

    @staticmethod
    def _calculate_null_ratio(data: pd.DataFrame, included_columns: List[str] = None):

        if not included_columns:
            return None

        na_columns = [col for col in included_columns if data[col].isnull().sum() > 0]

        ratio = (data[na_columns].isnull().sum() / data.shape[0])
        missing_df = pd.DataFrame(ratio)
        missing_df.reset_index(inplace=True)
        missing_df.columns = ["columns", "ratio"]
        return missing_df

    @staticmethod
    def _find_reduced_columns(missing_df: pd.DataFrame, null_ratio_threshold: float) -> List[str]:
        reduced_columns = [col for col in missing_df["columns"] if (missing_df[missing_df["columns"] == col]["ratio"]
                                                                    > null_ratio_threshold).any(axis=None)]
        return reduced_columns

    def get_description(self) -> str:
        return "Null Feature Reduction"



"""Includes DqComparisonColumnInfo class."""


class DqComparisonColumnInfo:
    """Dq comparison column settings."""

    def __init__(self):
        self.column_name: str = None
        self.benchmark_horizon: int = None
        self.duplicate_column_control: bool = None



"""Includes LassoSelectionOutput class"""
from organon.ml.feature_selection.domain.objects.base_selection_output import BaseSelectionOutput


class LassoSelectionOutput(BaseSelectionOutput):
    """Output of Lasso selection"""



"""This module includes "SumTriple" class."""
import numpy as np


class SumTriple:
    """Class for SumTriple"""

    def __init__(self, count: float = 0.0, sum_: float = 0.0, sum_of_squares: float = 0.0):
        self.count: float = count
        self.sum_: float = sum_
        self.sum_of_squares: float = sum_of_squares

    @property
    def average(self) -> float:
        """Returns average of SumTriple."""
        average = self.sum_ / self.count if self.count > 0 else float('nan')
        return average

    @property
    def variance(self) -> float:
        """Returns variance of SumTriple."""
        average = self.average
        variance = self.sum_of_squares / self.count - average * average if self.count > 0 else float('nan')
        return variance

    @property
    def std_dev(self) -> float:
        """Returns std_dev of SumTriple."""
        return np.sqrt(self.variance)

    def update_with_data(self, count: float, data: float):
        """Update class attribute with data"""
        self.count = self.count + count
        self.sum_ = self.sum_ + count * data
        self.sum_of_squares = self.sum_of_squares + count * data * data

    def update_with_sum_triple(self, triple:"SumTriple"):
        """Update class attribute with SumTriple Object"""
        self.count = self.count + triple.count
        self.sum_ = self.sum_ + triple.sum_
        self.sum_of_squares = self.sum_of_squares + triple.sum_of_squares
